import logging
import os
import sys

import numpy as np
import pandas as pd
import xarray

from pyaerocom import const
from pyaerocom.exceptions import (
    DataDimensionError,
    DataUnitError,
    EarlinetFileError,
    VarNotAvailableError,
)
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData
from pyaerocom.units.helpers import get_standard_unit
from pyaerocom.units.units_helpers import get_unit_conversion_fac
from pyaerocom.variable import Variable
from pyaerocom.vertical_profile import VerticalProfile

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)


class ReadEarlinet(ReadUngriddedBase):
    """Interface for reading of EARLINET data"""

    #: Mask for identifying datafiles
    _FILEMASK = "*.*"

    #: version log of this class (for caching)
    __version__ = "0.18_" + ReadUngriddedBase.__baseversion__

    #: Name of dataset (OBS_ID)
    DATA_ID = const.EARLINET_NAME

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [const.EARLINET_NAME]

    #: default variables for read method
    DEFAULT_VARS = ["bsc532aer", "ec532aer"]

    # These are applied to all files. If new cloud filter names are discovered they should be added here.
    # As of 20.08.24, however, there are still files with less reliable data which get through the filters.
    # https://github.com/metno/pyaerocom/issues/1310
    CLOUD_FILTERS = {
        "cloud_mask_type": 0,  # "no_cloudmask_available manual_cloudmask automatic_cloudmask"
        "cirrus_contamination": 2,  # "not_available no_cirrus cirrus_detected"
    }

    #: all data values that exceed this number will be set to NaN on read. This
    #: is because iris, xarray, etc. assign a FILL VALUE of the order of e36
    #: to missing data in the netcdf files
    _MAX_VAL_NAN = 1e6

    #: variable name of altitude in files
    ALTITUDE_ID = "altitude"

    #: temporal resolution
    # Note: This is an approximation based on the fact that MOST of the data appears to be collected
    # at an hourly reoslution. Some files are a little less, but typically this is the case
    TS_TYPE = "hourly"

    VAR_PATTERNS_FILE = {
        "ec532aer": "_e0532",
        "ec355aer": "_e0355",
        "bsc532aer": "_b0532",
        "bsc355aer": "_b0355",
        "bsc1064aer": "_b1064",
        # "zdust": "*.e*", # not sure if EARLINET has this anymore
    }

    #: dictionary specifying the file column names (values) for each Aerocom
    #: variable (keys)
    VAR_NAMES_FILE = {
        "ec532aer": "extinction",
        "ec355aer": "extinction",
        "ec1064aer": "extinction",
        "bsc532aer": "backscatter",
        "bsc355aer": "backscatter",
        "bsc1064aer": "backscatter",
        "zdust": "DustLayerHeight",  # not sure if EARLINET has this anymore
    }

    META_NAMES_FILE = dict(
        location="location",
        start_utc="measurement_start_datetime",
        stop_utc="measurement_stop_datetime",
        # wavelength_det="DetectionWavelength_nm",
        # res_raw_m="ResolutionRaw_meter",
        instrument_name="system",
        comment="comment",
        PI="PI",
        dataset_name="title",
        # station_name="station_ID",
        website="references",
        wavelength_emis="wavelength",
        # detection_mode="DetectionMode",
        # res_eval="ResolutionEvaluated",
        # input_params="InputParameters",
        altitude="altitude",
        # eval_method="backscatter_evaluation_method",
    )
    #: metadata keys that are needed for reading (must be values in
    #: :attr:`META_NAMES_FILE`)
    META_NEEDED = [
        "location",
        "measurement_start_datetime",
        "measurement_start_datetime",
    ]

    #: Metadata keys from :attr:`META_NAMES_FILE` that are additional to
    #: standard keys defined in :class:`StationMetaData` and that are supposed
    #: to be inserted into :class:`UngriddedData` object created in :func:`read`
    KEEP_ADD_META = [
        "location",
        "wavelength",
        "zenith_angle",
        "comment",
        "shots",
        "backscatter_evaluation_method",
    ]

    #: Attribute access names for unit reading of variable data
    VAR_UNIT_NAMES = dict(
        extinction=["units"],
        backscatter=["units"],
        dustlayerheight=["units"],
        altitude="units",
    )
    #: Variable names of uncertainty data
    ERR_VARNAMES = dict(
        ec532aer="error_extinction",
        ec355aer="error_extinction",
    )

    #: If true, the uncertainties are also read (where available, cf. ERR_VARNAMES)
    READ_ERR = True

    PROVIDES_VARIABLES = list(VAR_PATTERNS_FILE)

    EXCLUDE_CASES = ["cirrus.txt"]

    def __init__(self, data_id=None, data_dir=None):
        # initiate base class
        super().__init__(data_id=data_id, data_dir=data_dir)
        # make sure everything is properly set up
        if not all(
            [x in self.VAR_PATTERNS_FILE for x in self.PROVIDES_VARIABLES]
        ):  # pragma: no cover
            raise AttributeError(
                "Please specify file search masks in "
                "header dict VAR_PATTERNS_FILE for each "
                "variable defined in PROVIDES_VARIABLES"
            )
        elif not all(
            [x in self.VAR_NAMES_FILE for x in self.PROVIDES_VARIABLES]
        ):  # pragma: no cover
            raise AttributeError(
                "Please specify file search masks in "
                "header dict VAR_NAMES_FILE for each "
                "variable defined in PROVIDES_VARIABLES"
            )
        #: private dictionary containing loaded Variable instances,
        self._var_info = {}

        #: files that are supposed to be excluded from reading
        self.exclude_files = []

        #: files that were actually excluded from reading
        self.excluded_files = []

        self.is_vertical_profile = True

    @override
    def read_file(self, filename, vars_to_retrieve=None, read_err=None, remove_outliers=True):
        """Read EARLINET file and return it as instance of :class:`StationData`

        Parameters
        ----------
        filename : str
            absolute path to filename to read
        vars_to_retrieve : :obj:`list`, optional
            list of str with variable names to read. If None, use
            :attr:`DEFAULT_VARS`
        read_err : bool
            if True, uncertainty data is also read (where available).
        remove_outliers : bool
            if True, outliers are removed for each variable using the
            `minimum` and `maximum` attributes for that variable (accessed
            via pyaerocom.const.VARS[var_name]).

        Returns
        -------
        StationData
            dict-like object containing results
        """
        if read_err is None:  # use default setting
            read_err = self.READ_ERR
        if isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        _vars = []
        for var in vars_to_retrieve:
            if (
                var in self.VAR_PATTERNS_FILE
            ):  # make sure to only read what is supported by this file
                if self.VAR_PATTERNS_FILE[var] in filename:
                    _vars.append(var)
            elif var in self.AUX_REQUIRES:
                _vars.append(var)
            else:
                raise VarNotAvailableError(f"{var} is not supported")

        # implemented in base class
        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(_vars)

        # create empty data object (is dictionary with extended functionality)
        data_out = StationData()
        data_out["station_id"] = filename.split("/")[-1].split("_")[
            2
        ]  # loss of generality but should work. can also get from reading file if needed: data_in.station_ID
        data_out["data_id"] = self.data_id
        data_out["ts_type"] = self.TS_TYPE

        # create empty arrays for all variables that are supposed to be read
        # from file
        for var in vars_to_read:
            if var not in self._var_info:
                self._var_info[var] = Variable(var)
        var_info = self._var_info

        # Iterate over the lines of the file
        logger.debug(f"Reading file {filename}")

        with xarray.open_dataset(filename, engine="netcdf4", decode_timedelta=True) as data_in:
            for filter in self.CLOUD_FILTERS:
                if filter in data_in.variables:
                    if data_in.variables[filter].item() == self.CLOUD_FILTERS[filter]:
                        logger.debug(f"Skipping {filename} due to cloud filtering")
                        continue

            # getting the coords since no longer in metadata
            # Put also just in the attributes. not sure why appears twice
            data_out["station_coords"]["longitude"] = data_out["longitude"] = np.float64(
                data_in["longitude"].values
            )
            data_out["station_coords"]["latitude"] = data_out["latitude"] = np.float64(
                data_in["latitude"].values
            )
            data_out["altitude"] = np.float64(
                data_in[
                    "altitude"
                ].values  # altitude is defined in EARLINET in terms of altitude above sea level
            )  # Note altitude is an array for the data, station altitude is different

            data_out["station_coords"]["altitude"] = data_in.station_altitude.item()
            data_out["altitude_attrs"] = data_in[
                "altitude"
            ].attrs  # get attrs for altitude units + extra

            # get intersection of metadaa in ddataa_out and data_in
            for k, v in self.META_NAMES_FILE.items():
                if v in self.META_NEEDED:
                    _meta = data_in.attrs[v]
                else:
                    try:
                        _meta = data_in.attrs[v]
                    except Exception:  # pragma: no cover
                        _meta = None
                data_out[k] = _meta

            # get metadata expected in StationData but not in data_in's metadata
            data_out["wavelength_emis"] = data_in["wavelength"]
            data_out["shots"] = data_in["shots"].item()
            data_out["zenith_angle"] = np.float64(data_in["zenith_angle"])
            data_out["filename"] = filename
            if "Lev02" in filename:
                data_out["data_level"] = 2
            loc_split = data_in.attrs["location"].split(", ")
            data_out["station_name"] = loc_split[0]
            if len(loc_split) > 1:
                data_out["country"] = loc_split[1]

            dtime = (
                pd.Timestamp(data_in.measurement_start_datetime).to_numpy().astype("datetime64[s]")
            )
            stop = (
                pd.Timestamp(data_in.measurement_stop_datetime).to_numpy().astype("datetime64[s]")
            )

            # in case measurement goes over midnight into a new day
            if stop < dtime:
                stop = stop + np.timedelta64(1, "[D]")

            data_out["dtime"] = [dtime]
            data_out["stopdtime"] = [stop]
            data_out["has_zdust"] = False

            for var in vars_to_read:
                data_out["var_info"][var] = {}
                err_read = False
                unit_ok = False
                outliers_removed = False
                has_altitude = False

                netcdf_var_name = self.VAR_NAMES_FILE[var]
                # check if the desired variable is in the file
                if netcdf_var_name not in data_in.variables:
                    logger.warning(f"Variable {var} not found in file {filename}")
                    continue

                info = var_info[var]
                # xarray.DataArray
                arr = data_in.variables[netcdf_var_name]
                # the actual data as numpy array (or float if 0-D data, e.g. zdust)
                val = np.squeeze(np.float64(arr))  # squeeze to 1D array

                # CONVERT UNIT
                unit = None

                unames = self.VAR_UNIT_NAMES[netcdf_var_name]
                for u in unames:
                    if u in arr.attrs:
                        unit = arr.attrs[u]
                if unit is None:
                    raise DataUnitError(f"Unit of {var} could not be accessed in file {filename}")
                unit_fac = None
                try:
                    to_unit = self._var_info[var].units
                    unit_fac = get_unit_conversion_fac(unit, to_unit)
                    val *= unit_fac
                    unit = to_unit
                    unit_ok = True
                except Exception as e:
                    logger.warning(
                        f"Failed to convert unit of {var} in file {filename} (Earlinet): "
                        f"Error: {repr(e)}"
                    )

                # import errors if applicable
                err = np.nan
                if read_err and var in self.ERR_VARNAMES:
                    err_name = self.ERR_VARNAMES[var]
                    if err_name in data_in.variables:
                        err = np.squeeze(np.float64(data_in.variables[err_name]))
                        if unit_ok:
                            err *= unit_fac
                        err_read = True

                # 1D variable
                if var == "zdust":
                    if not val.ndim == 0:
                        raise DataDimensionError(
                            "Fatal: dust layer height data must be single value"
                        )

                    if unit_ok and info.minimum < val < info.maximum:
                        logger.warning(f"zdust value {val} out of range, setting to NaN")
                        val = np.nan

                    if np.isnan(val):
                        logger.warning(
                            f"Invalid value of variable zdust in file {filename}. Skipping...!"
                        )
                        continue

                    data_out["has_zdust"] = True
                    data_out[var] = val

                else:
                    if not val.ndim == 1:
                        raise DataDimensionError("Extinction data must be one dimensional")
                    elif len(val) == 0:
                        continue  # no data
                    # Remove NaN equivalent values
                    val[val > self._MAX_VAL_NAN] = np.nan

                    wvlg = var_info[var].wavelength_nm
                    wvlg_str = self.META_NAMES_FILE["wavelength_emis"]

                    assert data_in[wvlg_str].shape == (1,)
                    if not wvlg == float(data_in[wvlg_str][0]):
                        logger.info("No wavelength match")
                        continue

                    alt_id = self.ALTITUDE_ID
                    alt_data = data_in.variables[alt_id]

                    alt_vals = np.float64(alt_data)
                    alt_unit = alt_data.attrs[self.VAR_UNIT_NAMES[alt_id]]
                    to_alt_unit = get_standard_unit("alt")
                    if not alt_unit == to_alt_unit:
                        try:
                            alt_unit_fac = get_unit_conversion_fac(alt_unit, to_alt_unit)
                            alt_vals *= alt_unit_fac
                            alt_unit = to_alt_unit
                        except Exception as e:
                            logger.warning(f"Failed to convert unit: {repr(e)}")
                    has_altitude = True

                    # remove outliers from data, if applicable
                    if remove_outliers and unit_ok:
                        # REMOVE OUTLIERS
                        outlier_mask = np.logical_or(val < info.minimum, val > info.maximum)
                        val[outlier_mask] = np.nan

                        if err_read:
                            err[outlier_mask] = np.nan
                        outliers_removed = True
                    # remove outliers from errors if applicable
                    if err_read:
                        err[err > self._MAX_VAL_NAN] = np.nan

                    # create instance of ProfileData
                    profile = VerticalProfile(
                        data=val,
                        altitude=alt_vals,
                        dtime=dtime,
                        var_name=var,
                        data_err=err,
                        var_unit=unit,
                        altitude_unit=alt_unit,
                    )

                    # Write everything into profile
                    data_out[var] = profile

                data_out["var_info"][var].update(
                    unit_ok=unit_ok,
                    err_read=err_read,
                    outliers_removed=outliers_removed,
                    has_altitude=has_altitude,
                )
        return data_out

    @override
    def read(
        self,
        vars_to_retrieve=None,
        files=None,
        first_file=None,
        last_file=None,
        read_err=None,
        remove_outliers=True,
        pattern=None,
    ):
        """Method that reads list of files as instance of :class:`UngriddedData`

        Parameters
        ----------
        vars_to_retrieve : :obj:`list` or similar, optional,
            list containing variable IDs that are supposed to be read. If None,
            all variables in :attr:`PROVIDES_VARIABLES` are loaded
        files : :obj:`list`, optional
            list of files to be read. If None, then the file list is used that
            is returned on :func:`get_file_list`.
        first_file : :obj:`int`, optional
            index of first file in file list to read. If None, the very first
            file in the list is used
        last_file : :obj:`int`, optional
            index of last file in list to read. If None, the very last file
            in the list is used
        read_err : bool
            if True, uncertainty data is also read (where available). If
            unspecified (None), then the default is used (cf. :attr:`READ_ERR`)
         pattern : str, optional
            string pattern for file search (cf :func:`get_file_list`)

        Returns
        -------
        UngriddedData
            data object
        """

        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]

        if read_err is None:
            read_err = self.READ_ERR

        if files is None:
            if len(self.files) == 0:
                self.get_file_list(vars_to_retrieve, pattern=pattern)
            files = self.files

        # turn files into a list because I suspect there may be a bug if you don't do this
        if isinstance(files, str):
            files = [files]

        if first_file is None:
            first_file = 0
        if last_file is None:
            last_file = len(files)

        files = files[
            first_file : last_file + 1
        ]  # think need to +1 here in order to actually get desired subset

        self.read_failed = []

        data_obj = UngriddedData()
        data_obj.is_vertical_profile = True
        col_idx = data_obj.index
        meta_key = -1.0
        idx = 0

        # assign metadata object
        metadata = data_obj.metadata
        meta_idx = data_obj.meta_idx

        # last_station_id = ''
        num_files = len(files)

        disp_each = int(num_files * 0.1)
        if disp_each < 1:
            disp_each = 1

        VAR_IDX = -1
        for i, _file in enumerate(files):
            if i % disp_each == 0:
                logger.info(f"Reading file {i + 1} of {num_files} ({type(self).__name__})")
            try:
                stat = self.read_file(
                    _file,
                    vars_to_retrieve=vars_to_retrieve,
                    read_err=read_err,
                    remove_outliers=remove_outliers,
                )
                if not any([var in stat.vars_available for var in vars_to_retrieve]):
                    logger.info(
                        f"Station {stat.station_name} contains none of the desired variables. Skipping station..."
                    )
                    continue
                # if last_station_id != station_id:
                meta_key += 1
                # Fill the metadata dict
                # the location in the data set is time step dependant!
                # use the lat location here since we have to choose one location
                # in the time series plot
                metadata[meta_key] = {}
                metadata[meta_key].update(stat.get_meta())
                for add_meta in self.KEEP_ADD_META:
                    if add_meta in stat:
                        metadata[meta_key][add_meta] = stat[add_meta]
                # metadata[meta_key]['station_id'] = station_id

                metadata[meta_key]["data_revision"] = self.data_revision
                metadata[meta_key]["variables"] = []
                metadata[meta_key]["var_info"] = {}
                # this is a list with indices of this station for each variable
                # not sure yet, if we really need that or if it speeds up things
                meta_idx[meta_key] = {}
                # last_station_id = station_id

                # Is floating point single value
                time = stat.dtime[0]
                for var in stat.vars_available:
                    if var not in data_obj.var_idx:
                        VAR_IDX += 1
                        data_obj.var_idx[var] = VAR_IDX

                    var_idx = data_obj.var_idx[var]

                    val = stat[var]
                    metadata[meta_key]["var_info"][var] = vi = {}
                    if isinstance(val, VerticalProfile):
                        altitude = val.altitude
                        data = val.data
                        add = len(data)
                        err = val.data_err
                        metadata[meta_key]["var_info"]["altitude"] = via = {}

                        vi.update(val.var_info[var])
                        via.update(val.var_info["altitude"])
                    else:
                        add = 1
                        altitude = np.nan
                        data = val
                        if var in stat.data_err:
                            err = stat.err[var]
                        else:
                            err = np.nan
                    vi.update(stat.var_info[var])
                    stop = idx + add
                    # check if size of data object needs to be extended
                    if stop >= data_obj._ROWNO:
                        # if totnum < data_obj._CHUNKSIZE, then the latter is used
                        data_obj.add_chunk(add)

                    # write common meta info for this station
                    data_obj._data[idx:stop, col_idx["latitude"]] = stat["station_coords"][
                        "latitude"
                    ]
                    data_obj._data[idx:stop, col_idx["longitude"]] = stat["station_coords"][
                        "longitude"
                    ]
                    data_obj._data[idx:stop, col_idx["altitude"]] = stat["station_coords"][
                        "altitude"
                    ]
                    data_obj._data[idx:stop, col_idx["meta"]] = meta_key

                    # write data to data object
                    data_obj._data[idx:stop, col_idx["time"]] = time
                    data_obj._data[idx:stop, col_idx["stoptime"]] = stat.stopdtime[0]
                    data_obj._data[idx:stop, col_idx["data"]] = data
                    data_obj._data[idx:stop, col_idx["dataaltitude"]] = altitude
                    data_obj._data[idx:stop, col_idx["varidx"]] = var_idx

                    if read_err:
                        data_obj._data[idx:stop, col_idx["dataerr"]] = err

                    if var not in meta_idx[meta_key]:
                        meta_idx[meta_key][var] = []
                    meta_idx[meta_key][var].extend(list(range(idx, stop)))

                    if var not in metadata[meta_key]["variables"]:
                        metadata[meta_key]["variables"].append(var)

                    idx += add

            except Exception as e:
                self.read_failed.append(_file)
                logger.exception(f"Failed to read file {os.path.basename(_file)} (ERR: {repr(e)})")

        # shorten data_obj._data to the right number of points
        data_obj._data = data_obj._data[:idx]

        return data_obj

    def _get_exclude_filelist(self):  # pragma: no cover
        """Get list of filenames that are supposed to be ignored"""
        exclude = []
        import glob

        files = glob.glob(f"{self.data_dir}/EXCLUDE/*.txt")
        for i, file in enumerate(files):
            if os.path.basename(file) not in self.EXCLUDE_CASES:
                continue
            count = 0
            num = None
            indata = False
            with open(file) as f:
                for line in f:
                    if indata:
                        exclude.append(line.strip())
                        count += 1
                    elif "Number of" in line:
                        num = int(line.split(":")[1].strip())
                        indata = True

            if not count == num:
                raise EarlinetFileError(
                    "Number of excluded files does not match the number of files in the list"
                )
        self.exclude_files = list(dict.fromkeys(exclude))
        return self.exclude_files

    @override
    def get_file_list(self, vars_to_retrieve=None, pattern=None):
        """Perform recursive file search for all input variables

        Note
        ----
        Overloaded implementation of base class, since for Earlinet, the
        paths are variable dependent

        Parameters
        ----------
        vars_to_retrieve : list
            list of variables to retrieve
        pattern : str, optional
            file name pattern applied to search

        Returns
        -------
        list
            list containing file paths
        """

        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        exclude = self._get_exclude_filelist()
        logger.info("Fetching EARLINET data files. This might take a while...")
        patterns = []
        for var in vars_to_retrieve:
            if var not in self.VAR_PATTERNS_FILE:
                raise VarNotAvailableError(f"Input variable {var} is not supported")

            _pattern = self.VAR_PATTERNS_FILE[var]
            if pattern is not None:
                if "." in pattern:
                    raise NotImplementedError("filetype delimiter . not supported")
                spl = _pattern.split(".")
                if "*" not in spl[0]:
                    raise AttributeError(f"Invalid file pattern: {_pattern}")
                spl[0] = spl[0].replace("*", pattern)
                _pattern = ".".join(spl)

            patterns.append(_pattern)

        matches = []
        for root, dirnames, files in os.walk(self.data_dir, topdown=True):
            paths = [os.path.join(root, f) for f in files]
            for _pattern in patterns:
                for path in paths:
                    file = os.path.basename(path)
                    if _pattern not in file:
                        continue
                    elif file in exclude:
                        self.excluded_files.append(path)
                    else:
                        matches.append(path)
        self.files = files = list(dict.fromkeys(matches))
        return files



import logging
import os
import sys
from datetime import timedelta
from functools import cached_property
from pathlib import Path
from typing import Annotated

from pyaerocom.aeroval.glob_defaults import VarWebInfo, VarWebScaleAndColormap
from pyaerocom.aeroval.obsentry import ObsEntry

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import subprocess

import aerovaldb
import pandas as pd
from pydantic import (
    BaseModel,
    ConfigDict,
    computed_field,
    field_serializer,
    model_validator,
)

from pyaerocom.aeroval.aux_io_helpers import ReadAuxHandler
from pyaerocom.aeroval.collections import ModelCollection, ObsCollection
from pyaerocom.aeroval.exceptions import ConfigError
from pyaerocom.aeroval.helpers import (
    _check_statistics_periods,
    _get_min_max_year_periods,
    check_if_year,
)
from pyaerocom.aeroval.json_utils import read_json
from pyaerocom.colocation.colocation_setup import ColocationSetup

from .output_paths import OutputPaths
from .model_maps_setup import ModelMapsSetup
from .statistics_setup import StatisticsSetup
from .time_setup import TimeSetup
from .web_display_setup import WebDisplaySetup
from .eval_run_options import EvalRunOptions
from .project_info import ProjectInfo
from .experiment_info import ExperimentInfo
from .cams2_83_setup import CAMS2_83Setup
from .units_setup import UnitsSetup

logger = logging.getLogger(__name__)


class EvalSetup(BaseModel):
    """Composite class representing a whole analysis setup

    This represents the level at which json I/O happens for configuration
    setup files.
    """

    ###########################
    ##   Pydantic ConfigDict
    ###########################
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow", protected_namespaces=())

    ########################################
    ## Regular & BaseModel-based Attributes
    ########################################

    io_aux_file: Annotated[
        Path | str, ".py file containing additional read methods for modeldata"
    ] = ""

    var_web_info_file: Annotated[Path | str, "config file containing additional variables"] = ""

    var_scale_colmap_file: Annotated[
        Path | str, "config file containing scales/ranges for variables"
    ] = ""

    _aux_funs: dict = {}

    @model_validator(mode="after")
    def model_validator(self) -> Self:
        # Add missing variables to var_order_menu.
        var_order_menu = set(self.webdisp_opts.var_order_menu)
        obs_cfg = self.obs_cfg

        variables = set()
        for entry in obs_cfg:
            for var in entry.obs_vars:
                variables.add(var)

        if not var_order_menu.issuperset(variables):
            missing_vars = sorted(variables - var_order_menu)
            ls = list(self.webdisp_opts.var_order_menu)
            ls.extend(missing_vars)
            self.webdisp_opts.var_order_menu = tuple(ls)
            logger.info(
                f"Some variables are configured as obsvars but not included in var_order_menu. They have been appended to var_order_menu. Missing variables: {', '.join(missing_vars)}."
            )

        return self

    @computed_field
    @cached_property
    def proj_info(self) -> ProjectInfo:
        if not hasattr(self, "model_extra") or self.model_extra is None:
            return ProjectInfo()
        model_args = {
            key: val for key, val in self.model_extra.items() if key in ProjectInfo.model_fields
        }
        return ProjectInfo(**model_args)

    @computed_field
    @cached_property
    def exp_info(self) -> ExperimentInfo:
        model_args = {
            key: val for key, val in self.model_extra.items() if key in ExperimentInfo.model_fields
        }
        return ExperimentInfo(**model_args)

    @computed_field
    @cached_property
    def json_filename(self) -> str:
        """
        str: Savename of config file: cfg_<proj_id>_<exp_id>.json
        """
        return f"cfg_{self.proj_info.proj_id}_{self.exp_info.exp_id}.json"

    @cached_property
    def gridded_aux_funs(self) -> dict:
        if not bool(self._aux_funs) and os.path.exists(self.io_aux_file):
            self._import_aux_funs()
        return self._aux_funs

    @cached_property
    def var_web_info(self) -> VarWebInfo:
        return VarWebInfo(config_file=self.var_web_info_file)

    @cached_property
    def var_scale_colmap(self) -> VarWebScaleAndColormap:
        return VarWebScaleAndColormap(config_file=self.var_scale_colmap_file)

    @computed_field
    @cached_property
    def path_manager(self) -> OutputPaths:
        if not hasattr(self, "model_extra") or self.model_extra is None:
            return OutputPaths()
        model_args = {
            key: val for key, val in self.model_extra.items() if key in OutputPaths.model_fields
        }
        return OutputPaths(**model_args)

    # Many computed_fields here have this hack to get keys from a general CFG into their appropriate respective classes
    # TODO: all these computed fields could be more easily defined if the config were
    # rigid enough to have them explicitly defined (e.g., in a TOML file), rather than dumping everything
    # into one large config dict and then dishing out the relevant parts to each class.
    @computed_field
    @cached_property
    def time_cfg(self) -> TimeSetup:
        if not hasattr(self, "model_extra") or self.model_extra is None:
            return TimeSetup()
        model_args = {
            key: val for key, val in self.model_extra.items() if key in TimeSetup.model_fields
        }
        return TimeSetup(**model_args)

    @computed_field
    @cached_property
    def modelmaps_opts(self) -> ModelMapsSetup:
        if not hasattr(self, "model_extra") or self.model_extra is None:
            return ModelMapsSetup()
        model_args = {
            key: val for key, val in self.model_extra.items() if key in ModelMapsSetup.model_fields
        }
        return ModelMapsSetup(**model_args)

    @computed_field
    @cached_property
    def cams2_83_cfg(self) -> CAMS2_83Setup:
        if not hasattr(self, "model_extra"):
            return CAMS2_83Setup()
        model_args = {
            key: val for key, val in self.model_extra.items() if key in CAMS2_83Setup.model_fields
        }
        return CAMS2_83Setup(**model_args)

    @computed_field
    @cached_property
    def webdisp_opts(self) -> WebDisplaySetup:
        if not hasattr(self, "model_extra") or self.model_extra is None:
            return WebDisplaySetup()
        model_args = {
            key: val
            for key, val in self.model_extra.items()
            if key in WebDisplaySetup.model_fields
        }
        return WebDisplaySetup(**model_args)

    @computed_field
    @cached_property
    def processing_opts(self) -> EvalRunOptions:
        if not hasattr(self, "model_extra") or self.model_extra is None:
            return EvalRunOptions()
        model_args = {
            key: val for key, val in self.model_extra.items() if key in EvalRunOptions.model_fields
        }
        return EvalRunOptions(**model_args)

    @computed_field
    @cached_property
    def statistics_opts(self) -> StatisticsSetup:
        if not hasattr(self, "model_extra") or self.model_extra is None:
            return StatisticsSetup(weighted_stats=True, annual_stats_constrained=False)
        model_args = {
            key: val
            for key, val in self.model_extra.items()
            if key in StatisticsSetup.model_fields
        }
        return StatisticsSetup(**model_args)

    @computed_field
    @cached_property
    def colocation_opts(self) -> ColocationSetup:
        if not hasattr(self, "model_extra") or self.model_extra is None:
            return ColocationSetup(save_coldata=True, keep_data=False, resample_how="mean")

        model_args = {
            key: val
            for key, val in self.model_extra.items()
            if key in ColocationSetup.model_fields
        }
        # need to pass some default values to the ColocationSetup if not provided in config
        default_dict = {
            "save_coldata": True,
            "keep_data": False,
            "resample_how": "mean",
        }
        for key in default_dict:
            if key not in model_args:
                model_args[key] = default_dict[key]

        return ColocationSetup(**model_args)

    ##################################
    ## Non-BaseModel-based attributes
    ##################################

    # These attributes require special attention b/c they're not based on Pydantic's BaseModel class.

    @computed_field
    @cached_property
    def obs_cfg(self) -> ObsCollection:
        oc = ObsCollection()
        for k, v in self.model_extra.get("obs_cfg", {}).items():
            oc.add_entry(k, v)
        return oc

    @field_serializer("obs_cfg")
    def serialize_obs_cfg(self, obs_cfg: ObsCollection):
        return obs_cfg.as_dict()

    @computed_field
    @cached_property
    def model_cfg(self) -> ModelCollection:
        mc = ModelCollection()
        for k, v in self.model_extra.get("model_cfg", {}).items():
            mc.add_entry(k, v)
        return mc

    @computed_field
    @cached_property
    def units_cfg(self) -> UnitsSetup:
        if not hasattr(self, "model_extra") or self.model_extra is None:
            return UnitsSetup()
        model_args = {
            key: val for key, val in self.model_extra.items() if key in UnitsSetup.model_fields
        }
        return UnitsSetup(**model_args)

    @field_serializer("model_cfg")
    def serialize_model_cfg(self, model_cfg: ModelCollection):
        return model_cfg.as_dict()

    @computed_field
    @cached_property
    def pip_freeze(self) -> list[str]:
        reqs = subprocess.check_output([sys.executable, "-m", "pip", "freeze"])
        splt = [x for x in reqs.decode().split("\n") if x != ""]

        return sorted(splt)

    ###########################
    ##       Methods
    ###########################

    def get_obs_entry(self, obs_name) -> ObsEntry:
        """Returns ObsEntry instance for network obs_name"""
        return self.obs_cfg.get_entry(obs_name)

    def get_model_entry(self, model_name) -> dict:
        """Get model entry configuration

        Since the configuration files for experiments are in json format, they
        do not allow the storage of executable custom methods for model data
        reading. Instead, these can be specified in a python module that may
        be specified via :attr:`add_methods_file` and that contains a
        dictionary `FUNS` that maps the method names with the callable methods.

        As a result, this means that, by default, custom read methods for
        individual models in :attr:`model_config` do not contain the
        callable methods but only the names. This method will take care of
        handling this and will return a dictionary where potential custom
        method strings have been converted to the corresponding callable
        methods.

        Parameters
        ----------
        model_name : str
            name of model

        Returns
        -------
        dict
            Dictionary that specifies the model setup ready for the analysis
        """
        cfg = self.model_cfg.get_entry(model_name)
        cfg = cfg.prep_dict_analysis(self.gridded_aux_funs)
        return cfg

    def to_json(self, outdir: str, ignore_nan: bool = True, indent: int = 3) -> None:
        """
        Save configuration as JSON file

        Parameters
        ----------
        outdir : str
            directory where the config json file is supposed to be stored
        ignore_nan : bool
            set NaNs to Null when writing
        indent : int
            json indentation

        """
        with aerovaldb.open(
            self.path_manager.json_basedir
            if self.path_manager.avdb_resource is None
            else self.path_manager.json_basedir
        ) as db:
            with db.lock():
                db.put_config(self.json_repr(), self.proj_info.proj_id, self.exp_info.exp_id)

    @staticmethod
    def from_json(filepath: str) -> Self:
        """Load configuration from json config file"""
        settings = read_json(filepath)
        return EvalSetup(**settings)

    def json_repr(self):
        return self.model_dump()

    def _import_aux_funs(self) -> None:
        h = ReadAuxHandler(self.io_aux_file)
        self._aux_funs.update(**h.import_all())

    def _check_time_config(self) -> None:
        periods = self.time_cfg.periods
        colstart = self.colocation_opts.start
        colstop = self.colocation_opts.stop

        if len(periods) == 0:
            if colstart is None:
                raise ConfigError("Either periods or start must be set...")
            per = self.colocation_opts._period_from_start_stop()
            periods = [per]
            logger.info(
                f"periods is not set, inferred {per} from start / stop colocation settings."
            )

        self.time_cfg.periods = _check_statistics_periods(periods)
        start, stop = _get_min_max_year_periods(periods)
        start_yr = start.year
        stop_yr = stop.year
        years = check_if_year(periods)
        if not years:
            if start == stop and isinstance(start, pd.Timestamp):
                stop = start + timedelta(hours=23)
            elif isinstance(start, pd.Timestamp):
                stop = stop + timedelta(hours=23)

            if stop_yr == start_yr:
                stop_yr += 1
            if colstart is None:
                self.colocation_opts.start = start.strftime("%Y/%m/%d %H:%M:%S")
            if colstop is None:
                self.colocation_opts.stop = stop.strftime(
                    "%Y/%m/%d %H:%M:%S"
                )  # + 1  # add 1 year since we want to include stop year
        else:
            if colstart is None:
                self.colocation_opts.start = start_yr
            if colstop is None:
                self.colocation_opts.stop = (
                    stop_yr + 1
                )  # add 1 year since we want to include stop year



from copy import deepcopy
import pytest
import numpy as np
from pathlib import Path

from pyaerocom.aeroval.bulkfraction_engine import BulkFractionEngine
from pyaerocom.aeroval import ExperimentProcessor
from pyaerocom.aeroval._processing_base import HasColocator, ProcessingEngine, ExperimentOutput
from pyaerocom.aeroval import EvalSetup
from pyaerocom import ColocatedData
from tests.fixtures.aeroval import cfg_test_bulk


@pytest.fixture
def cfg(tmpdir) -> dict:
    cfg = deepcopy(cfg_test_bulk.CFG)
    cfg["json_basedir"] = f"{tmpdir}/data"
    cfg["coldata_basedir"] = f"{tmpdir}/coldata"
    return cfg


@pytest.fixture
def bulkengine_instance(cfg: dict) -> BulkFractionEngine:
    setup = deepcopy(EvalSetup(**cfg))
    bfe = BulkFractionEngine(setup)
    return bfe


def test___init__(cfg: dict):
    setup = EvalSetup(**cfg)
    bfe = BulkFractionEngine(setup)
    assert isinstance(bfe, ProcessingEngine)
    assert isinstance(bfe, HasColocator)


def test__get_bulk_vars(bulkengine_instance: BulkFractionEngine):
    obsentry = bulkengine_instance.cfg.obs_cfg.get_entry("EBAS-m")
    assert (
        bulkengine_instance._get_bulk_vars("concwetrdn", obsentry)
        == obsentry.bulk_options["concwetrdn"].vars
    )

    with pytest.raises(KeyError, match="Could not find bulk vars entry"):
        bulkengine_instance._get_bulk_vars("concwetrdn2", obsentry)


def test_get_colocators_modelexists(bulkengine_instance: BulkFractionEngine):
    obsentry = bulkengine_instance.cfg.obs_cfg.get_entry("EBAS-m")
    freq = obsentry.ts_type
    bulk_vars = bulkengine_instance._get_bulk_vars("concwetrdn", obsentry)
    cols_exist = bulkengine_instance.get_colocators(
        bulk_vars, "concwetrdn", freq, "EMEP", "EBAS-m", True
    )

    assert len(cols_exist) == 2

    assert (
        cols_exist[0][bulk_vars[0]].colocation_setup.model_use_vars[bulk_vars[0]]
        == cols_exist[1][bulk_vars[1]].colocation_setup.model_use_vars[bulk_vars[1]]
    )


def test_get_colocators(bulkengine_instance: BulkFractionEngine):
    obsentry = bulkengine_instance.cfg.obs_cfg.get_entry("EBAS-m")
    freq = obsentry.ts_type
    bulk_vars = bulkengine_instance._get_bulk_vars("concwetrdn", obsentry)
    cols_exist = bulkengine_instance.get_colocators(
        bulk_vars, "concwetrdn", freq, "EMEP", "EBAS-m", False
    )

    assert len(cols_exist) == 2

    assert cols_exist[0][bulk_vars[0]].colocation_setup.obs_vars[0] == bulk_vars[0]
    assert cols_exist[1][bulk_vars[1]].colocation_setup.obs_vars[0] == bulk_vars[1]


def test__run_var(bulkengine_instance: BulkFractionEngine):
    obs_name = "AERONET-Sun"
    model_name = "TM5-AP3-CTRL"
    var_name = "fraction"
    freq = "monthly"
    obsentry = bulkengine_instance.cfg.obs_cfg.get_entry(obs_name)
    modelentry = bulkengine_instance.cfg.model_cfg.get_entry(model_name)
    bulk_vars = bulkengine_instance._get_bulk_vars(var_name, obsentry)

    col, fp = bulkengine_instance._run_var(
        model_name, obs_name, var_name, bulk_vars, freq, obsentry, modelentry
    )

    assert isinstance(fp[0], str)
    assert isinstance(col, ColocatedData)
    assert pytest.approx(np.nanmean(col.data), rel=1e-5) == 1.0


def test__combine_coldatas(bulkengine_instance: BulkFractionEngine):
    obs_name = "AERONET-Sun"
    model_name = "TM5-AP3-CTRL"
    var_name = "fraction"
    freq = "monthly"
    obsentry = bulkengine_instance.cfg.obs_cfg.get_entry(obs_name)
    modelentry = bulkengine_instance.cfg.model_cfg.get_entry(model_name)
    bulk_vars = bulkengine_instance._get_bulk_vars(var_name, obsentry)

    num_name = bulk_vars[0]
    denum_name = bulk_vars[1]

    model_exists = obsentry.bulk_options[var_name].model_exists
    cols = bulkengine_instance.get_colocators(
        bulk_vars, var_name, freq, model_name, obs_name, model_exists
    )

    coldatas = []
    for col in cols:
        if len(list(col.keys())) != 1:
            raise ValueError(
                "Found more than one colocated object when trying to run bulk variable"
            )
        bv = list(col.keys())[0]
        coldatas.append(col[bv].run(bv))

    obsentry2 = obsentry.model_copy()
    obsentry2.bulk_options[var_name].mode = "fraction"
    model_num_name, model_denum_name = bulkengine_instance._get_model_var_names(
        var_name, bulk_vars, model_exists, modelentry
    )

    data1 = bulkengine_instance._combine_coldatas(
        coldatas[0][model_num_name][num_name],
        coldatas[1][model_denum_name][denum_name],
        var_name,
        obsentry2,
    )

    assert pytest.approx(np.nanmean(data1.data), rel=1e-5) == 1.0
    assert len(data1.data) == max(
        len(coldatas[0][model_num_name][num_name].data),
        len(coldatas[1][model_denum_name][denum_name].data),
    )

    obsentry3 = obsentry.model_copy()
    obsentry3.bulk_options[var_name].mode = "product"
    data2 = bulkengine_instance._combine_coldatas(
        coldatas[0][model_num_name][num_name],
        coldatas[1][model_denum_name][denum_name],
        var_name,
        obsentry3,
    )

    assert len(data2.data) == max(
        len(coldatas[0][model_num_name][num_name].data),
        len(coldatas[1][model_denum_name][denum_name].data),
    )
    assert np.array_equal(
        data2.data.data, coldatas[0][model_num_name][num_name].data.data ** 2, equal_nan=True
    )


def test__combine_coldatas_model_exist(bulkengine_instance: BulkFractionEngine):
    obs_name = "AERONET-Sun-exist"
    model_name = "TM5-AP3-CTRL"
    var_name = "abs550aer"
    freq = "monthly"

    obsentry = bulkengine_instance.cfg.obs_cfg.get_entry(obs_name)
    modelentry = bulkengine_instance.cfg.model_cfg.get_entry(model_name)
    bulk_vars = bulkengine_instance._get_bulk_vars(var_name, obsentry)

    num_name = bulk_vars[0]
    denum_name = bulk_vars[1]

    model_exists = obsentry.bulk_options[var_name].model_exists
    cols = bulkengine_instance.get_colocators(
        bulk_vars, var_name, freq, model_name, obs_name, model_exists
    )

    coldatas = []
    for col in cols:
        if len(list(col.keys())) != 1:
            raise ValueError(
                "Found more than one colocated object when trying to run bulk variable"
            )
        bv = list(col.keys())[0]
        coldatas.append(col[bv].run(bv))

    obsentry2 = obsentry.model_copy()
    obsentry2.bulk_options[var_name].mode = "fraction"
    model_num_name, model_denum_name = bulkengine_instance._get_model_var_names(
        var_name, bulk_vars, model_exists, modelentry
    )

    data1 = bulkengine_instance._combine_coldatas(
        coldatas[0][model_num_name][num_name],
        coldatas[1][model_denum_name][denum_name],
        var_name,
        obsentry2,
    )

    assert pytest.approx(np.nanmean(data1.data[0]), rel=1e-5) == 1.0
    assert pytest.approx(np.nanmean(data1.data[1]), rel=1e-5) != 0.0135


def test_run(bulkengine_instance: BulkFractionEngine):
    obs_name = "AERONET-Sun-exist"
    model_name = "TM5-AP3-CTRL"
    var_name = "abs550aer"

    bulkengine_instance.exp_output.delete_experiment_data(also_coldata=True)

    bulkengine_instance.run(var_name, model_name, obs_name)

    output: ExperimentOutput = bulkengine_instance.exp_output
    assert Path(output.exp_dir).is_dir()


def test_run_cfg(cfg: dict):
    model_name = "TM5-AP3-CTRL"
    setup = EvalSetup(**cfg)
    proc = ExperimentProcessor(setup)
    proc.exp_output.delete_experiment_data(also_coldata=True)
    proc.run(obs_name="AERONET-Sun", model_name="TM5-AP3-CTRL")

    output: ExperimentOutput = proc.exp_output
    assert Path(output.exp_dir).is_dir()

    output.avdb.get_experiments(output.proj_id)
    data = output.avdb.get_timeseries(
        output.proj_id, output.exp_id, "ALL", "AERONET-Sun", "fraction", "Column"
    )
    m_data = data[model_name]["monthly_mod"]
    o_data = data[model_name]["monthly_obs"]
    assert pytest.approx(np.nanmean(m_data), rel=1e-5) == 1.0
    assert pytest.approx(np.nanmean(o_data), rel=1e-5) == 1.0



from __future__ import annotations

import getpass
import logging
import os.path
from pathlib import Path

import pytest

import pyaerocom.config as testmod
from pyaerocom import const
from pyaerocom.config import ALL_REGION_NAME, Config
from pyaerocom.data import resources
from pyaerocom.grid_io import GridIO
from pyaerocom.varcollection import VarCollection
from pyaerocom.variable import Variable
from tests.conftest import lustre_avail

USER = getpass.getuser()

with resources.path("pyaerocom.data", "paths.ini") as path:
    DEFAULT_PATHS_INI = str(path)

logger = logging.getLogger(__name__)


@pytest.fixture()
def config_file(file: str | None, tmp_path: Path) -> str | None:
    if file is None:
        return None

    if file.casefold() == "default":
        assert resources.is_resource("pyaerocom.data", "paths.ini")
        with resources.path("pyaerocom.data", "paths.ini") as path:
            return str(path)

    if file.casefold() == "wrong_suffix":
        path = tmp_path / "paths.txt"
        path.write_text("")
        assert path.exists()
        return str(path)

    return file


@pytest.fixture()
def local_db(tmp_path: Path) -> Path:
    """temporary path to DB file structure"""
    path = tmp_path / "data"
    (path / "modeldata").mkdir(parents=True)
    (path / "obsdata").mkdir()
    assert path.is_dir()
    return path


@pytest.fixture(scope="module")
def empty_cfg():
    cfg = testmod.Config(try_infer_environment=False)
    return cfg


@pytest.mark.parametrize(
    "file,try_infer_environment",
    [
        (None, False),
        (None, True),
        ("default", False),
    ],
)
def test_Config___init__(config_file: str, try_infer_environment: bool):
    testmod.Config(config_file, try_infer_environment)


@pytest.mark.parametrize(
    "file,exception,error",
    [
        pytest.param(
            "wrong_suffix",
            ValueError,
            "Need path to an ini file for input config_file",
            id="wrong file extension",
        ),
        pytest.param(
            f"/home/{USER}/blaaaa.ini",
            FileNotFoundError,
            f"input config file does not exist /home/{USER}/blaaaa.ini",
            id="no such file",
        ),
    ],
)
def test_Config___init___error(config_file: str, exception: type[Exception], error: str):
    with pytest.raises(exception) as e:
        testmod.Config(config_file, False)
    assert str(e.value) == error


def test_Config_has_access_lustre():
    cfg = testmod.Config(try_infer_environment=False)
    assert not cfg.has_access_lustre


def test_user_specific_paths_ini():
    # test if user specific paths.ini file is read
    CHANGE_NAME = "NAME_CHANGED_FOR_TESTING"
    CHECK_NAME = "GAWTADSUBSETAASETAL"
    user_file = os.path.join(const.my_pyaerocom_dir, const.PATHS_INI_NAME)
    # only create user_file if it doesn't exist
    del_flag = False
    if not os.path.exists(user_file):
        with open(DEFAULT_PATHS_INI) as infile, open(user_file, "w") as outfile:
            for line in infile:
                if CHECK_NAME in line:
                    line = f"{CHECK_NAME} = {CHANGE_NAME}\n"
                else:
                    line = line.replace("/lustre/storeB/project", "${HOME}")
                outfile.write(line)
        del_flag = True

    assert os.path.exists(user_file)
    # no real test here for now since we would need to get rid of the already loaded const module
    # and recreate that The following does not work due to caching
    # cfg = testmod.Config(try_infer_environment=False)
    # assert cfg.GAWTADSUBSETAASETAL == CHANGE_NAME

    if del_flag:
        os.remove(user_file)


def test_Config_read_config():
    cfg = testmod.Config(try_infer_environment=False)
    cfg_file = DEFAULT_PATHS_INI
    assert Path(cfg_file).exists()
    cfg.read_config(cfg_file)
    # not all paths from the default paths.ini are present on CI
    # Just test a few of them
    assert Path(cfg.OUTPUTDIR).exists()
    assert Path(cfg.COLOCATEDDATADIR).exists()
    assert Path(cfg.CACHEDIR).exists()


def test_empty_class_header(empty_cfg):
    cfg = empty_cfg
    assert cfg.AERONET_SUN_V3L15_AOD_DAILY_NAME == "AeronetSunV3Lev1.5.daily"
    assert cfg.AERONET_SUN_V3L15_AOD_ALL_POINTS_NAME == "AeronetSunV3Lev1.5.AP"
    assert cfg.AERONET_SUN_V3L2_AOD_DAILY_NAME == "AeronetSunV3Lev2.daily"
    assert cfg.AERONET_SUN_V3L2_AOD_ALL_POINTS_NAME == "AeronetSunV3Lev2.AP"
    assert cfg.AERONET_SUN_V3L15_SDA_DAILY_NAME == "AeronetSDAV3Lev1.5.daily"
    assert cfg.AERONET_SUN_V3L15_SDA_ALL_POINTS_NAME == "AeronetSDAV3Lev1.5.AP"
    assert cfg.AERONET_SUN_V3L2_SDA_DAILY_NAME == "AeronetSDAV3Lev2.daily"
    assert cfg.AERONET_SUN_V3L2_SDA_ALL_POINTS_NAME == "AeronetSDAV3Lev2.AP"
    assert cfg.AERONET_INV_V3L15_DAILY_NAME == "AeronetInvV3Lev1.5.daily"
    assert cfg.AERONET_INV_V3L2_DAILY_NAME == "AeronetInvV3Lev2.daily"
    assert cfg.EBAS_MULTICOLUMN_NAME == "EBASMC"
    assert cfg.EEA_NAME == "EEAAQeRep"
    assert cfg.EARLINET_NAME == "EARLINET"
    assert cfg.GAWTADSUBSETAASETAL_NAME == "GAWTADsubsetAasEtAl"
    assert cfg.DMS_AMS_CVO_NAME == "DMS_AMS_CVO"
    assert cfg.CNEMC_NAME == "CNEMC"
    assert cfg.EBAS_DB_LOCAL_CACHE
    assert cfg.MIN_YEAR == 0
    assert cfg.MAX_YEAR == 20000
    assert cfg.STANDARD_COORD_NAMES == ["latitude", "longitude", "altitude"]
    assert cfg.DEFAULT_VERT_GRID_DEF == dict(lower=0, upper=15000, step=250)
    assert cfg.RH_MAX_PERCENT_DRY == 40
    assert cfg.DEFAULT_REG_FILTER == f"{ALL_REGION_NAME}-wMOUNTAINS"
    assert cfg.OBS_MIN_NUM_RESAMPLE == dict(
        yearly=dict(monthly=3),
        monthly=dict(daily=7),
        daily=dict(hourly=6),
        hourly=dict(minutely=15),
    )
    from pyaerocom import obs_io

    assert cfg.OBS_ALLOW_ALT_WAVELENGTHS == obs_io.OBS_ALLOW_ALT_WAVELENGTHS

    assert cfg.OBS_WAVELENGTH_TOL_NM == obs_io.OBS_WAVELENGTH_TOL_NM

    assert cfg.CLIM_START == 2005
    assert cfg.CLIM_STOP == 2015
    assert cfg.CLIM_FREQ == "monthly"
    assert cfg.CLIM_RESAMPLE_HOW == "mean"
    assert cfg.CLIM_MIN_COUNT == dict(daily=30, monthly=5)

    # names for the satellite data sets
    assert cfg.SENTINEL5P_NAME == "Sentinel5P"
    assert cfg.AEOLUS_NAME == "AeolusL2A"

    assert cfg.OLD_AEROCOM_REGIONS == [
        ALL_REGION_NAME,
        "ASIA",
        "AUSTRALIA",
        "CHINA",
        "EUROPE",
        "INDIA",
        "NAFRICA",
        "SAFRICA",
        "SAMERICA",
        "NAMERICA",
    ]

    assert cfg.URL_HTAP_MASKS == "https://pyaerocom.met.no/pyaerocom-suppl/htap_masks/"

    assert cfg.HTAP_REGIONS == [
        "PAN",
        "EAS",
        "NAF",
        "MDE",
        "LAND",
        "SAS",
        "SPO",
        "OCN",
        "SEA",
        "RBU",
        "EEUROPE",
        "NAM",
        "WEUROPE",
        "SAF",
        "USA",
        "SAM",
        "EUR",
        "NPO",
        "MCA",
    ]

    assert cfg.RM_CACHE_OUTDATED

    #: Name of the file containing the revision string of an obs data network
    assert cfg.REVISION_FILE == "Revision.txt"

    assert cfg._outhomename == "MyPyaerocom"

    with resources.path("pyaerocom.data", "variables.ini") as path:
        assert cfg._var_info_file == str(path)
    with resources.path("pyaerocom.data", "coords.ini") as path:
        assert cfg._coords_info_file == str(path)

    assert cfg.DO_NOT_CACHE_FILE is None

    assert cfg.ERA5_SURFTEMP_FILENAME == "era5.msl.t2m.201001-201012.nc"

    assert cfg._LUSTRE_CHECK_PATH == "/"


def test_empty_init(empty_cfg):
    cfg = empty_cfg
    assert cfg._cache_basedir is None
    assert cfg._outputdir is None
    assert cfg._colocateddatadir is None
    assert cfg._filtermaskdir is None
    assert cfg._local_tmp_dir is None
    assert cfg._downloaddatadir is None
    assert cfg._confirmed_access == []
    assert cfg._rejected_access == []

    # Options
    assert cfg._caching_active is True

    assert cfg._var_param is None
    assert cfg._coords is None

    # Attributes that are used to store search directories
    assert cfg.OBSLOCS_UNGRIDDED == {}
    assert cfg.OBS_UNGRIDDED_POST == {}
    assert cfg.SUPPLDIRS == {}
    assert cfg._search_dirs == []

    assert cfg.WRITE_FILEIO_ERR_LOG is True

    assert cfg.last_config_file is None
    assert cfg._ebas_flag_info is None
    from pyaerocom.grid_io import GridIO

    #: Settings for reading and writing of gridded data
    assert isinstance(cfg.GRID_IO, GridIO)


def test_default_config_HOMEDIR():
    assert Path(const.HOMEDIR) == Path("~").expanduser()
    assert const.HOMEDIR.endswith("/")


@lustre_avail
def test_default_config():
    cfg = Config()

    mypydir = Path(cfg.HOMEDIR).resolve() / "MyPyaerocom"
    assert Path(cfg.OUTPUTDIR) == Path(cfg._outputdir) == mypydir
    assert Path(cfg.CACHEDIR) == mypydir / f"_cache/{USER}"
    assert Path(cfg.COLOCATEDDATADIR) == Path(cfg._colocateddatadir) == mypydir / "colocated_data"
    assert Path(cfg.FILTERMASKKDIR) == Path(cfg._filtermaskdir) == mypydir / "filtermasks"
    assert Path(cfg.LOCAL_TMP_DIR) == Path(cfg._local_tmp_dir) == mypydir / "tmp"
    assert Path(cfg.DOWNLOAD_DATADIR) == Path(cfg._downloaddatadir) == mypydir / "data"

    assert cfg._caching_active
    assert cfg.CACHING

    assert isinstance(cfg.VARS, VarCollection)
    assert cfg.VARS is cfg.VAR_PARAM
    assert isinstance(cfg._var_param, VarCollection)

    assert isinstance(cfg.COORDINFO, VarCollection)
    assert cfg._coords is cfg.COORDINFO

    assert cfg.DATA_SEARCH_DIRS is cfg._search_dirs

    assert cfg.WRITE_FILEIO_ERR_LOG

    assert isinstance(cfg.GRID_IO, GridIO)


def test_register_variable_with_dict():
    test_var_name = "conctestvariabledict"
    variables = {
        test_var_name: {
            "var_name": test_var_name,
            "units": "ug m-3",
        }
    }
    const.register_custom_variables(variables)

    vars = const.VARS

    assert test_var_name in vars.find(test_var_name)


def test_register_variable_with_Variable():
    test_var_name = "testvariableVariable"
    variables = {
        test_var_name: Variable(
            var_name=test_var_name,
            units="ug m-3",
        ),
    }
    const.register_custom_variables(variables)

    vars = const.VARS

    assert test_var_name in vars.all_vars



import pathlib
import xarray as xr
import os
import logging
from pyaerocom.units import Unit
from pyaerocom.io.uemep import uemep_variables
from pyaerocom.colocation.colocated_data import validate_structure
from pyaerocom.io.readungridded import ReadUngridded
import pyaerocom
import time
import pandas as pd
import datetime


from pyaerocom.units.units_helpers import get_unit_conversion_fac


logger = logging.getLogger(__name__)


class UEMEPColocator:
    """
    Helper class for colcating uEMEP station data with observation
    data.

    Limitations:

    - Only colocates hourly model data with hourly obs data. No temporal
    resampling occurs.
    - Only works for EBAS data currently.

    :param uemep_station_data : Path to folder containing uemep_station_data, or list of files. Files
    must be readable using xarray's open_mfdataset().
    :param obs : dict or list with observations to be colocated against. If list, must be
    data_ids understood by ReadUngridded. If dict, must be a mapping from identifier to data_id
    understood by ReadUngridded. The identifier can be chosen freely and is only used for metadata
    (which is shown on Aeroval).
    :var_names : A list of variables. If not provided, all variables defined in uemep_variables.toml
    will be tried.
    :out_dir : Directory where colocated data objects will be written. Note that output will be one
    file per variable/observation combination. Variables will be stored as files following aerocom
    convention.
    """

    # Maps an aerocom variable name to it's equivalent uemep variable name.
    VAR_LOOKUP = uemep_variables()

    def __init__(
        self,
        uemep_station_data: os.PathLike | str | list[os.PathLike | str],
        *,
        obs: dict[str, str] | list[str],
        var_names: str | list[str] | None = None,
        out_dir: os.PathLike | None = None,
    ):
        """
        :param uemep_station_data: Folder with .nc UEMEP station data.
        :param var_names: List of aerocom variable names to colocate.
        :param obs: Either list of obs_ids passed to ReadUngridded, or dict with mapping
        of identifier to obs_id passed to ReadUngridded.
        :param out_dir: Path to output directory where colocated data objects will be stored.
        Defaults to '.'.
        """
        if isinstance(var_names, str):
            var_names = [var_names]

        if not isinstance(uemep_station_data, list):
            logger.info("Looking for station data in '%s'.", uemep_station_data)
            self._file_path = list(pathlib.Path(uemep_station_data).glob("*.nc"))
        else:
            logger.info("Using the following files as station data: '%s'", uemep_station_data)
            self._file_path = [pathlib.Path(p) for p in uemep_station_data]

        self._uemep_station_data = None

        if var_names is None:
            var_names = list(UEMEPColocator.VAR_LOOKUP.keys())
            logger.info(
                "No variable list configured. Defaulting to all configured uemep variables: %s",
                var_names,
            )

        aerocom_vars = []
        for var in var_names:
            if var not in UEMEPColocator.VAR_LOOKUP.keys():
                logger.warning(
                    "Variable '%s' not found in configured uEMEP variables. It will be ignored.",
                    var,
                )
                continue

            aerocom_vars.append(var)

        self._vars = sorted(set(aerocom_vars))

        if isinstance(obs, list):
            self._obs = {v: ReadUngridded(v) for v in obs}
        elif isinstance(obs, dict):
            self._obs = {k: ReadUngridded(v) for k, v in obs.items()}
        else:
            raise TypeError

        if out_dir is None:
            out_dir = pathlib.Path(".")

        self._out_dir = out_dir

    def _load_station_data(self) -> None:
        logger.info("Reading uEMEP station data. This may take a while.")
        start_time = time.perf_counter()
        with xr.open_mfdataset(self._file_path, engine="netcdf4", decode_timedelta=True) as dt:
            self._uemep_station_data = dt
        logger.info(f"Finished reading data in {time.perf_counter() - start_time:.3f} seconds.")

    @property
    def uemep_station_data(self) -> xr.Dataset:
        if self._uemep_station_data is None:
            self._load_station_data()

        return self._uemep_station_data

    def _run_single_variable(self, var: str) -> None:
        uemep_name = UEMEPColocator.VAR_LOOKUP[var]
        logger.info("Using uemep variable '%s' for aerocom variable '%s'", uemep_name, var)

        uemep_data = self.uemep_station_data[uemep_name].swap_dims({"station_id": "station_name"})
        model_unit = Unit(uemep_data.attrs["units"])
        uemep_data = uemep_data.assign_coords(
            {"station_name": uemep_data.station_name.astype(str)}
        ).assign_coords({"time": uemep_data.time + pd.Timedelta(minutes=30)})
        station_ids = uemep_data["station_name"].values.astype(str)

        for obs_id, obs_reader in self._obs.items():
            logger.info("Running colocation against obs_id '%s'.", obs_id)

            obsdata = obs_reader.read(vars_to_retrieve=[var])

            fobsdata = obsdata.filter_by_meta(ts_type="hourly")

            start_date = str(uemep_data.time.min().values)
            end_date = str(uemep_data.time.max().values)

            logger.info("uEMEP data spans %s-%s", start_date, end_date)

            sdata: dict[str, pyaerocom.stationdata.StationData] = {}
            stations = fobsdata.to_station_data_all()
            for _, station in zip(stations["station_name"], stations["stats"]):
                ids = station.station_id.split(";")
                if not any([id in station_ids for id in ids]):
                    logger.info(
                        "Station '%s' not found in uemep data. Skipping.",
                        station.station_id,
                    )
                    continue

                sdata[station.station_id] = station

            if len(sdata.keys()) == 0:
                logger.error(
                    "No matching stations found in '%s'. Aborting colocation of '%s'.",
                    obs_id,
                    var,
                )
                continue

            for s in station_ids:  # Remove timeseries for stations not present in model data.
                if s not in sdata.keys():
                    uemep_data = uemep_data.drop(s, dim="station_name")

            darrays: list[xr.DataArray] = []
            sids = []
            for sid, station in sdata.items():
                ts = station.to_timeseries(var).loc[start_date:end_date]
                if len(ts) == 0:
                    logger.warning("Length of timeseries for '%s' is 0.", station.station_id)
                    continue

                unit = Unit(station.var_info[var]["units"])

                conversion_factor = get_unit_conversion_fac(unit, model_unit)
                if conversion_factor != 1:
                    logger.info(
                        f"Transforming timeseries from '{unit}' to '{model_unit}' using conversion factor {float(conversion_factor):.2}."
                    )
                    ts *= conversion_factor

                darrays.append(
                    xr.DataArray(ts, dims=["time"], coords={"time": ts.index}, name=var)
                )
                sids.append(sid)

            if len(darrays) == 0:
                logger.error("No stations with data for the given time range.")
                return

            combined = xr.concat(darrays, dim=pd.Index(sids, name="station_name"))
            uemep_data = uemep_data.expand_dims(data_source=["uemep"])
            combined = combined.expand_dims(data_source=[obs_id])
            combined = combined.assign_coords(
                {"station_name": [x for x in combined.station_name.values]}
            )

            coldataarray = xr.concat([combined, uemep_data], dim="data_source")

            coldataarray = coldataarray.transpose("data_source", "time", "station_name").rename(
                {"lat": "latitude", "lon": "longitude"}
            )
            coldataarray = coldataarray.drop_vars("station_id")
            coldat = pyaerocom.colocation.colocated_data.ColocatedData(coldataarray)
            coldat.data.attrs = {
                "obs_vars": var,
                "ts_type": "hourly",
                "filter_name": "ALL-wMOUNTAINS",
                "ts_type_src": ["hourly", "hourly"],
                "var_units": [
                    uemep_data.attrs["units"],
                    uemep_data.attrs["units"],
                ],
                "data_level": 2,
                "revision_ref": datetime.datetime.strftime(datetime.date.today(), "%Y%m%d"),
                "from_files": [str(p) for p in self._file_path],
                "from_files_ref": [],
                "colocate_time": 0,  # ?
                "obs_is_clim": 0,
                "pyaerocom": pyaerocom.__version__,
                "CONV!min_num_obs": str(dict(monthly=dict(daily=3))),
                "resample_how": "mean",
                "obs_name": obs_id,
                "vert_code": "Surface",
                "diurnal_only": 0,
                "zeros_to_nan": 0,
                "data_source": [obs_id, "uemep"],
                "var_name": [var, var],
            }
            validate_structure(coldat.data)
            logger.info("Writing colocated data object to file in dir '%s'.", self._out_dir)
            coldat.to_netcdf(self._out_dir)

    def run(self) -> None:
        logger.info("Starting colocation.")
        for i, var in enumerate(self._vars, start=1):
            logger.info("Processing variable %d of %d: %s", i, len(self._vars), var)
            self._run_single_variable(var)

        logger.info("Finished.")


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG, handlers=[logging.StreamHandler()])
    colocator = UEMEPColocator(
        uemep_station_data=pathlib.Path(
            "/lustre/storeB/project/fou/kl/emep/ModelRuns/uEMEP/uEMEP_norway/rerun/2023/stations"
        ),
        obs=["EBASMC"],
        var_names=["conco3"],
    )
    colocator.run()



from __future__ import annotations

from datetime import date, datetime
from enum import Enum
from pathlib import Path
from typing import NamedTuple


class ModelName(str, Enum):
    ENSEMBLE = "ensemble"
    CHIMERE = "chimere"
    DEHM = "dehm"
    EMEP = "emep"
    EURAD = "euradim"
    GEMAQ = "gemaq"
    LOTOS = "lotos"
    MATCH = "match"
    MINNI = "minni"
    MOCAGE = "mocage"
    MONARCH = "monarch"
    SILAM = "silam"
    IFS = "ifs"

    def __str__(self) -> str:
        return self.value


class RunType(str, Enum):
    FC = "forecast"
    AN = "analysis"

    def __str__(self) -> str:
        return self.value

    @property
    def days(self) -> int:
        return dict(FC=3, AN=0)[self.name]


class ModelData(NamedTuple):
    name: ModelName
    run: RunType
    date: date = date.today()
    root: Path = Path.cwd()

    def __str__(self) -> str:
        return f"{self.name.name} {self.date:%F} {self.run.name}"

    @property
    def path(self) -> Path:
        return self.root / self.date.strftime(f"%Y%m/%Y%m%d_{self.name}_{self.run}.nc")

    @classmethod
    def frompath(cls, path: str | Path) -> ModelData:
        if isinstance(path, str):
            path = Path(path)
        date, name, run = path.stem.split("_")
        return cls(
            ModelName(name),
            RunType(run),
            datetime.strptime(date, "%Y%m%d").date(),
            path.parents[1],
        )



import numpy as np


class FilterNaN:
    """
    Excludes data for which either data or ref_data is NaN.
    """

    def __call__(
        self, data: np.ndarray, ref_data: np.ndarray, weights: np.ndarray | None
    ) -> np.ndarray:
        mask = ~np.isnan(ref_data) * ~np.isnan(data)

        return mask


class FilterByLimit:
    """
    Filters out data which is outside of a lower and / or upper limit.
    """

    def __init__(self, lowlim: float | None, highlim: float | None):
        self.lowlim = lowlim
        self.highlim = highlim

    def __call__(
        self, data: np.ndarray, ref_data: np.ndarray, weights: np.ndarray | None
    ) -> np.ndarray:
        valid1 = np.repeat([True], len(data))
        valid2 = np.repeat([True], len(data))

        if self.lowlim is not None:
            valid1 = np.logical_and(data > self.lowlim, ref_data > self.lowlim)
        if self.highlim is not None:
            valid2 = np.logical_and(data < self.highlim, ref_data < self.highlim)

        return valid1 * valid2



from pyaerocom.config import ALL_REGION_NAME

MODELS = {
    "TM5-AP3-CTRL": dict(
        model_id="TM5-met2010_CTRL-TEST",
        model_add_vars=dict(od550csaer=["od550aer"]),
        model_ts_type_read="monthly",
        flex_ts_type=False,
    )
}

ODCSFUN = "AeronetSDAV3L2Subset.daily;od550lt1aer+AeronetSDAV3L2Subset.daily;od550gt1aer"
OBS_GROUNDBASED = {
    "AERONET-Sun": dict(
        obs_id="AeronetSunV3L2Subset.daily", obs_vars=("od550aer",), obs_vert_type="Column"
    ),
    "AERONET-SDA": dict(
        obs_id="AERONET-SDA",
        obs_vars=["od550csaer"],
        obs_type="ungridded",
        obs_vert_type="Column",
        obs_merge_how={
            "od550csaer": "eval",
        },
        obs_aux_requires={
            "od550csaer": {
                "AeronetSDAV3L2Subset.daily": ["od550lt1aer", "od550gt1aer"],
            }
        },
        obs_aux_funs={"od550csaer": ODCSFUN},
        obs_aux_units={"od550csaer": "1"},
    ),
}

CFG = dict(
    model_cfg=MODELS,
    obs_cfg=OBS_GROUNDBASED,
    json_basedir="PATH_TO_AEROVAL_OUT/data",
    coldata_basedir="PATH_TO_AEROVAL_OUT/coldata",
    # if True, existing colocated data files will be deleted
    reanalyse_existing=True,
    raise_exceptions=True,
    only_json=False,
    add_model_maps=False,
    only_model_maps=False,
    clear_existing_json=False,
    # Regional filter for analysis
    filter_name=f"{ALL_REGION_NAME}-wMOUNTAINS",
    # colocation frequency (no statistics in higher resolution can be computed)
    ts_type="monthly",
    map_zoom="World",
    freqs=["monthly"],
    periods=["2010"],
    main_freq="monthly",
    zeros_to_nan=False,
    min_num_obs=None,
    colocate_time=False,
    obs_remove_outliers=False,
    add_seasons=False,
    model_remove_outliers=False,
    harmonise_units=True,
    regions_how="country",  #'default',#'country',
    annual_stats_constrained=False,
    proj_id="test",
    exp_id="exp2",
    exp_name="AeroVal test experiment 2",
    exp_descr=("Test setup for more complex evaluation configurations"),
    exp_pi="Jonas Gliss",
    public=True,
    # directory where colocated data files are supposed to be stored
    weighted_stats=True,
)



from __future__ import annotations

from pathlib import Path

import pytest

from pyaerocom.aeroval import ExperimentProcessor
from pyaerocom.aeroval.experiment_output import ExperimentOutput
from pyaerocom.aeroval import EvalSetup

CHK_CFG1 = {
    "map": ["AERONET-Sun_od550aer_Column_TM5-AP3-CTRL_od550aer_2010.json"],
    "contour": 0,
    "hm": [
        "glob_stats_daily.json",
        "glob_stats_monthly.json",
        "glob_stats_yearly.json",
    ],
    "hm/ts": 10,  # number of .json files in sub dir
    "scat": ["AERONET-Sun_od550aer_Column_TM5-AP3-CTRL_od550aer_2010.json"],
    "ts": 11,  # number of .json files in subdir
    "ts/diurnal": 0,  # number of .json files in subdir
}

CHK_CFG2 = {
    "map": [
        "AERONET-SDA_od550aer_Column_TM5-AP3-CTRL_od550aer_2010.json",
        "AERONET-Sun_od550aer_Column_TM5-AP3-CTRL_od550aer_2010.json",
    ],
    "contour": 0,
    "hm": ["glob_stats_monthly.json"],
    "hm/ts": 21,
    "scat": [
        "AERONET-SDA_od550aer_Column_TM5-AP3-CTRL_od550aer_2010.json",
        "AERONET-Sun_od550aer_Column_TM5-AP3-CTRL_od550aer_2010.json",
    ],
    "ts": 40,
    "ts/diurnal": 0,  # number of .json files in subdir
}

CHK_CFG4 = {
    "map": ["SDA-and-Sun_od550aer_Column_TM5-AP3-CTRL_od550aer_2010.json"],
    "contour": 0,
    "hm": ["glob_stats_monthly.json"],
    "hm/ts": 10,  # number of .json files in subdir
    "scat": ["SDA-and-Sun_od550aer_Column_TM5-AP3-CTRL_od550aer_2010.json"],
    "ts": 21,  # number of .json files in subdir
    "ts/diurnal": 0,  # number of .json files in subdir
}


@pytest.mark.parametrize(
    "cfg,chk_files",
    [
        ("cfgexp1", CHK_CFG1),
        ("cfgexp2", CHK_CFG2),
        ("cfgexp4", CHK_CFG4),
    ],
)
def test_ExperimentOutput__FILES(eval_config: dict, chk_files: dict):
    cfg = EvalSetup(**eval_config)
    proc = ExperimentProcessor(cfg)
    proc.exp_output.delete_experiment_data(also_coldata=True)
    proc.run()

    output: ExperimentOutput = proc.exp_output
    assert Path(output.exp_dir).is_dir()
    assert Path(output.experiments_file).exists()

    output.avdb.get_statistics(output.proj_id, output.exp_id)
    output.avdb.get_menu(output.proj_id, output.exp_id)
    output.avdb.get_ranges(output.proj_id, output.exp_id)

    output.avdb.get_config(output.proj_id, output.exp_id)

    for key, path in cfg.path_manager.get_json_output_dirs().items():
        path = Path(path)
        assert path.is_dir()
        if key not in chk_files:
            continue

        check = chk_files[key]
        if isinstance(check, list):
            files = [file.name for file in path.iterdir()]
            assert all(file in files for file in check)
        elif isinstance(check, int):
            files = list(path.glob("*.json"))
            assert len(files) == check


@pytest.mark.parametrize("cfg,reanalyse_existing", [("cfgexp4", True), ("cfgexp4", False)])
def test_reanalyse_existing(eval_config: dict, reanalyse_existing: bool):
    eval_config["reanalyse_existing"] = reanalyse_existing
    cfg = EvalSetup(**eval_config)
    assert cfg.colocation_opts.reanalyse_existing == reanalyse_existing
    output = Path(cfg.path_manager.coldata_basedir) / cfg.proj_id / cfg.exp_id

    proc = ExperimentProcessor(cfg)
    proc.exp_output.delete_experiment_data(also_coldata=True)
    assert not output.exists()

    proc.run()
    assert output.is_dir()
    assert list(output.glob("**/*.nc"))

    proc.exp_output.delete_experiment_data(also_coldata=False)
    assert output.is_dir()
    assert list(output.glob("**/*.nc"))


@pytest.mark.parametrize("cfg", ["cfgexp4"])
def test_superobs_different_resolutions(eval_config: dict):
    cfg = EvalSetup(**eval_config)
    cfg.model_cfg.get_entry("TM5-AP3-CTRL").model_ts_type_read = None
    cfg.model_cfg.get_entry("TM5-AP3-CTRL").flex_ts_type = True

    cfg.obs_cfg.get_entry("AERONET-Sun").ts_type = "daily"
    cfg.obs_cfg.get_entry("AERONET-SDA").ts_type = "monthly"

    proc = ExperimentProcessor(cfg)
    proc.exp_output.delete_experiment_data(also_coldata=True)
    proc.run()



from __future__ import annotations

from datetime import datetime, timedelta
from pathlib import Path
from textwrap import dedent

import pytest

from pyaerocom import const
from pyaerocom.io.cams2_83.obs import DEFAULT_METADATA_NAME, read_csv
from pyaerocom.io.cams2_83.read_obs import DATA_FOLDER_PATH, ReadCAMS2_83
from pyaerocom.io.cams2_83.read_obs import obs_paths as find_obs_paths
from pyaerocom.io.readungridded import ReadUngridded
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.ungridded_data_container import UngriddedDataContainer

TEST_DATE = datetime(2021, 12, 1)
TEST_DATES = [TEST_DATE + timedelta(days=d) for d in range(3)]


@pytest.fixture
def obs_paths() -> list[Path]:
    if not DATA_FOLDER_PATH.is_dir():
        pytest.skip(f"no access to {DATA_FOLDER_PATH}")
    paths = find_obs_paths(*TEST_DATES)
    return list(paths)


@pytest.fixture
def obs_file(tmp_path: Path) -> Path:
    obs = """
        STATION;LAT;LON;ALT(m);PARAMETER;YEAR;MONTH;DAY;HOUR;AVERAGING_PERIOD(h);CONCENTRATION(kg/m3)
        AT0ENK1;48.392; 13.671;0525;o3;2025;02;12;01;1; 4.27600e-08
        AT0ILL1;47.770; 16.766;0117;o3;2025;02;12;01;1; 6.56700e-08
        """

    path = tmp_path / "tmp_d1" / "obs.csv"
    path.parent.mkdir(parents=True)
    path.write_text(dedent(obs))
    return path


@pytest.fixture
def metadata_file(tmp_path: Path) -> Path:
    return tmp_path / DEFAULT_METADATA_NAME


def test_init():
    data = ReadCAMS2_83()
    assert isinstance(data, ReadUngriddedBase)


def test_obs_paths(obs_paths: list[Path]):
    for path in obs_paths:
        assert path.exists()


def test_read_ungridded(obs_paths: list[Path]):
    data = ReadUngridded().read(const.CAMS2_83_NRT_NAME, "concco", files=obs_paths)
    assert isinstance(data, UngriddedDataContainer)


def test_obs_no_metadata_file(
    obs_file: Path, metadata_file: Path, caplog: pytest.LogCaptureFixture
):
    df = read_csv(obs_file, polls=["O3"])
    assert f"Metadata file {metadata_file} does not exist" in caplog.text
    assert list(df) == ["station", "lat", "lon", "alt", "time", "poll", "conc"]


def test_obs_invalid_metadata_file(
    obs_file: Path, metadata_file: Path, caplog: pytest.LogCaptureFixture
):
    metadata = """
        something not parsable as expected
        bla;bla;bla
        bla;bla;bla
        """
    metadata_file.write_text(dedent(metadata))

    df = read_csv(obs_file, polls=["O3"])
    assert list(df) == ["station", "lat", "lon", "alt", "time", "poll", "conc"]
    assert "Invalid metadata file" in caplog.text


def test_obs_empty_metadata_file(
    obs_file: Path, metadata_file: Path, caplog: pytest.LogCaptureFixture
):
    metadata = "station, station_type"
    metadata_file.write_text(metadata)

    df = read_csv(obs_file, polls=["O3"])
    assert list(df) == ["station", "lat", "lon", "alt", "time", "poll", "conc"]
    assert "Empty metadata" in caplog.text


def test_obs_ok_metadata_file(obs_file: Path, metadata_file: Path):
    metadata = """
        something, something, something
        bla, bla, bla
        bla, bla, bla
        """
    metadata_file.write_text(dedent(metadata))

    df = read_csv(obs_file, polls=["O3"])
    assert list(df) == [
        "station",
        "lat",
        "lon",
        "alt",
        "time",
        "poll",
        "conc",
        "station_type",
    ]
    assert df["station_type"].isna().all()


def test_obs_read_to_ungridded(
    obs_file: Path, metadata_file: Path, caplog: pytest.LogCaptureFixture
):
    metadata = """
        something, something, something
        AT0ENK1, bla, rur
        AT0ILL1, bla, sub
        """
    metadata_file.write_text(dedent(metadata))

    reader = ReadCAMS2_83()
    data = reader.read(vars_to_retrieve=["conco3"], files=[obs_file])
    assert isinstance(data, UngriddedDataContainer)
    assert "Time needed to convert obs to ungridded" in caplog.text
    assert all("station_type" in dict for dict in data.metadata.values())
    assert {dict["station_type"] for dict in data.metadata.values()} == {"rur", "sub"}



from __future__ import annotations

from datetime import datetime, timedelta

import numpy as np
import pandas as pd

DEP_TYPE = {
    1: "Throughfall",
    2: "Bulk",
    3: "Wet-only",
    4: "Stemflow",
    5: "Fog",
    6: "Frozen fog",
    9: "Other",
    8: "do not use",
}

COUNTRY_CODES = {
    1: "FR",
    2: "BE",
    3: "NL",
    4: "DE",
    5: "IT",
    6: "UK",
    7: "IE",
    8: "DK",
    9: "GR",
    10: "PT",
    11: "ES",
    12: "LU",
    13: "SE",
    14: "AT",
    15: "FI",
    50: "CH",
    51: "HU",
    52: "RO",
    53: "PL",
    54: "SK",
    55: "NO",
    56: "LT",
    57: "HR",
    58: "CZ",
    59: "EE",
    60: "SI",
    61: "MD",
    62: "RU",
    63: "BG",
    64: "LV",
    65: "BY",
    66: "CY",
    67: "CS",
    68: "AD",
    95: "cn",
    80: "ME",
    96: "AZ",
    72: "TR",
}

COUNTRIES = {
    1: "France",
    2: "Belgium",
    3: "Netherlands",
    4: "Germany",
    5: "Italy",
    6: "United Kingdom",
    7: "Ireland",
    8: "Denmark",
    9: "Greece",
    10: "Portugal",
    11: "Spain",
    12: "Luxembourg",
    13: "Sweden",
    14: "Austria",
    15: "Finland",
    50: "Switzerland",
    51: "Hungary",
    52: "Romania",
    53: "Poland",
    54: "Slovak Republic",
    55: "Norway",
    56: "Lithuania",
    57: "Croatia",
    58: "Czech Republic",
    59: "Estonia",
    60: "Slovenia",
    61: "Republic of Moldova",
    62: "Russia",
    63: "Bulgaria",
    64: "Latvia",
    65: "Belarus",
    66: "Cyprus",
    67: "Serbia",
    68: "Andorra",
    95: "Canaries (Spain)",
    80: "Montenegro",
    96: "Azores (Portugal)",
    72: "Türkiye",
}


class Station:
    """Holds information for a singe ICP Forest station. The same station with different ts_types
    are treated as separate stations at this level

    Parameters
    ----------
    country_code : int
        int representing the country where the station is found
    plot_code : int
        The number of the station
    sampler_code : int
        int representing the type of sampler
    lat : str
        latitude
    lon : str
        longitude
    alt: int
        altitude
    partner_code : int
        number of the institute doing the measurement. Mostly used for metadata
    ts_type : str
        ts_type of the station

    Attributes
    ----------
    sampler_type : str
        The type of wetdep measurement
    data : dict[str, list[float]]
        dictionary holding a list of measurements for each species
    dtime : dict[str, list[datetime]]
        dictionary holding a list of timesteps for each species
    country : str
        name of country
    station_name : str
        name of station
    var_info : dict[str, dict[str, str | list[float]]]
        dict holder metadata for each species, used by StationData


    """

    def __init__(
        self,
        country_code: int,
        plot_code: int,
        sampler_code: int,
        lat: str,
        lon: str,
        alt: int,
        partner_code: int,
        ts_type: str,
    ) -> None:
        self.country_code = country_code
        self.plot_code = plot_code
        self.sampler_code = sampler_code
        self.lat = lat
        self.lon = lon
        self.alt = alt
        self.partner_code = partner_code
        self.ts_type = ts_type

        self.sampler_type = DEP_TYPE[self.sampler_code]

        self.data: dict[str, list[float]] = {}
        self.flags: dict[str, list[int]] = {}
        self.dtime: dict[str, list[datetime]] = {}

        self.country = COUNTRIES[country_code]

        self.station_name = self.get_station_name(country_code, plot_code, sampler_code)

        self.var_info: dict[str, dict[str, str | list[float]]] = {}

    @staticmethod
    def get_station_name(
        country_code: int,
        plot_code: int,
        sampler_code: int,
    ) -> str:
        """
        Generates the station name based on country code, plot code and sampler code. Static method
        that can be used without initiating the class

        Parameters
        ----------
        country_code : int
            int representing the country where the station is found
        plot_code : int
            The number of the station
        sampler_code : int
            int representing the type of sampler

        Returns
        -------
        str
            station name

        """
        return f"{COUNTRY_CODES[country_code]}-{plot_code}-{sampler_code}"

    def _add_species_to_var_info(self, species: str, unit: str) -> None:
        """
        Makes the var_info dict that is used later by StationData

        Parameters
        ----------
        species : str
            name of species
        unit : str
            unit

        """
        self.var_info[species] = dict(
            ts_type=self.ts_type,
            ts_type_src=self.ts_type,
            units=unit,
            sampler_type=self.sampler_type,
        )

    def add_measurement(
        self, species: str, time: datetime, measurement: float, unit: str, flag: int
    ) -> None:
        """
        Adds a single measurement to the data and time lists. If it is the first measurement,
        a new var_info is created

        Parameters
        ----------
        species : str
            name of species
        time : datetime
            The timestamp of the measurement
        measurement : float
            The value of the measurement
        unit : str
            unit

        """
        if species not in self.var_info:
            self._add_species_to_var_info(species, unit)
        if species not in self.dtime:
            self.dtime[species] = []
            self.data[species] = []
            self.flags[species] = []

        self.dtime[species].append(time)
        self.data[species].append(measurement)
        self.flags[species].append(flag)

    def get_timeseries(self, species: str, quality_limit: float = 0.5) -> pd.Series:
        """
        Combines the data and timestamps of a given species to a pandas timeseries

        Parameters
        ----------
         species : str
            name of species

        Returns
        -------
        pd.Series
            Timeseries of the measurements of a given species for station

        """
        flags = np.array(self.flags[species])
        quality = np.sum(flags[np.where(flags == 0)]) / len(flags)
        if quality >= quality_limit:
            return pd.Series(self.data[species], index=self.dtime[species])
        else:
            return pd.Series(np.ones_like(self.data[species]) * np.nan, index=self.dtime[species])


class SurveyYear:
    """
    Holds information about the start and stop dates, as well
    as the number of periods for a single year at a single station


    Parameters
    ----------
    year: int
        The year
    start: datetime
        start date
    stop: datetime
        end date
    periods: int
        number of periods in the survey year

    Attributes
    ----------
    days : float
        The number of days in each period
    ts_type : str
        The ts_type for the survey year


    """

    def __init__(self, year: int, start: datetime, stop: datetime, periods: int) -> None:
        self.year = year
        self.start = start
        self.stop = stop
        self.periods = periods

        # self.daterange = date_range(start, stop, periods)
        self.days = (self.stop - self.start).days / self.periods

        self.ts_type = self._get_tstype()

    def get_date(self, period: int) -> datetime:
        """
        Finds the date used to make each period in the year.
        For now the function returns the first date in the period

        Parameters
        ----------
        period : int
            Which period to find the date for

        Returns
        -------
        float
            The first date in the period

        Raises
        ------
        ValueError
            If the period is not within the range of periods in the year
        """
        if period > self.periods or period <= 0:
            raise ValueError(f"The period {period} needs to be in the range 1-{self.periods}")
        return self.start + timedelta(days=self.days * (period - 1))

    def _get_tstype(self) -> str:
        """
        Returns the ts_type by calling the static method

        Returns
        -------
        str
            the ts_type of the survey year
        """

        return SurveyYear.get_tstype(self.days)

    @staticmethod
    def get_tstype(days: float) -> str:
        if days >= 26:
            return "monthly"
        elif days >= 12:
            return "2weekly"
        elif days >= 6:
            return "weekly"
        else:
            return "daily"


class Plot:
    def __init__(
        self,
        country_code: int,
        plot_code: int,
        sampler_code: int,
        lat: str,
        lon: str,
        alt: int,
        partner_code: int,
    ) -> None:
        self.country_code = country_code
        self.plot_code = plot_code
        self.sampler_code = sampler_code
        self.lat = lat
        self.lon = lon
        self.alt = alt
        self.partner_code = partner_code

        self.periods = {}
        self.survey_years: dict[int, SurveyYear] = {}

    def add_survey_year(self, year: int, start: str, stop: str, periods: int) -> None:
        start_dt = datetime.strptime(start, "%Y-%m-%d")
        stop_dt = datetime.strptime(stop, "%Y-%m-%d")
        self.survey_years[year] = SurveyYear(year, start_dt, stop_dt, periods)

    def get_date(self, year: int, period: int) -> datetime:
        return self.survey_years[year].get_date(period)


class Plots:
    def __init__(self, plot_file: str) -> None:
        self.plot_file: str = plot_file
        self.plots: dict[int, dict[int, dict[int, Plot]]] = {}

    def read_file(self, altitudes: dict[str, int]) -> dict[int, dict[int, dict[int, Plot]]]:
        plots: dict[int, dict[int, dict[int, Plot]]] = {}
        print("Starting to read plot metadata")
        with open(self.plot_file) as f:
            f.readline()
            for line in f:
                words = line.split(";")
                if words[0] == "":
                    continue

                survey_year = int(words[0])
                country_code = int(words[1])
                partner_code = int(words[2])
                plot_code = int(words[3])
                sampler_code = int(words[4])

                lat = words[6]
                lon = words[7]
                alt_code = words[8]
                start = words[9]
                stop = words[10]
                periods = int(words[11])

                alt = altitudes[alt_code]

                if start == "" or stop == "":
                    continue

                if country_code not in plots:
                    plots[country_code] = {}
                if plot_code not in plots[country_code]:
                    plots[country_code][plot_code] = {}
                if sampler_code not in plots[country_code][plot_code]:
                    plots[country_code][plot_code][sampler_code] = Plot(
                        country_code, plot_code, sampler_code, lat, lon, alt, partner_code
                    )

                plots[country_code][plot_code][sampler_code].add_survey_year(
                    survey_year, start, stop, periods
                )
        print("Done read plot metadata")
        self.plots = plots
        return plots

    def get_ts_type(self, year: int, country_code: int, plot_code: int, sampler_code: int) -> str:
        return self.plots[country_code][plot_code][sampler_code].survey_years[year].ts_type

    def get_date(
        self, year: int, country_code: int, plot_code: int, sampler_code: int, period: int
    ) -> datetime:
        return self.plots[country_code][plot_code][sampler_code].get_date(year, period)

    def get_days(self, year: int, country_code: int, plot_code: int, sampler_code: int) -> float:
        start = self.plots[country_code][plot_code][sampler_code].survey_years[year].start
        stop = self.plots[country_code][plot_code][sampler_code].survey_years[year].stop
        days = self.plots[country_code][plot_code][sampler_code].survey_years[year].days

        if start == stop:
            raise ValueError(f"start {start} is the same as stop {stop}")

        return days

    def get_position(
        self, year: int, country_code: int, plot_code: int, sampler_code: int
    ) -> tuple[float, float, int]:
        lat = self._coord_to_desimal(self.plots[country_code][plot_code][sampler_code].lat)
        lon = self._coord_to_desimal(self.plots[country_code][plot_code][sampler_code].lon)
        alt = self.plots[country_code][plot_code][sampler_code].alt

        return lat, lon, alt

    def _coord_to_desimal(self, coord: str) -> float:
        sign = 1
        if "-" in coord:
            sign = -1
            coord = coord[1:]

        if coord == "0":
            return 0

        coord = coord[:-2]
        if len(coord) >= 2:
            minute = int(coord[-2:])
            coord = coord[:-2]
        else:
            return sign * (int(coord) / 60.0)

        degree = int(coord)

        return sign * (degree + (minute / 60.0))


class MetadataReader:
    def __init__(self, dir: str) -> None:
        self.dir = dir

        self.add_dir = dir + "/adds"

        self.altitudes = self._get_altitude_dir()

        self.plots = Plots(self.dir + "/dp_pld.csv")
        self.plots.read_file(self.altitudes)

        self.deposition_type = DEP_TYPE

    def _get_altitude_dir(self) -> dict[str, int]:
        altitudes = {}
        with open(self.add_dir + "/dictionaries/d_altitude.csv") as f:
            f.readline()
            for line in f:
                words = line.split(";")
                altitudes[words[0]] = int(words[4]) + (int(words[5]) - int(words[4])) // 2

        return altitudes



"""
Compatibility layer between importlib.resources for Python 3.11 and older versions
"""

import sys
from contextlib import AbstractContextManager
from pathlib import Path

if sys.version_info >= (3, 11):
    from importlib import resources
else:
    import importlib_resources as resources


def path(package: str, resource: str) -> AbstractContextManager[Path]:
    """A context manager providing a file path object to the resource.
    If the resource does not already exist on its own on the file system,
    e.g. it only exists in a zip-file, the resource will be extracted and
    a temporary file will be created. If the file was created, the file
    will be deleted upon exiting the context manager (no exception is
    raised if the file was deleted prior to the context manager
    exiting).
    """
    return resources.as_file(resources.files(package) / resource)


def is_resource(package: str, name: str) -> bool:
    """True if `name` is a resource inside `package`.

    Directories are *not* resources.
    """
    with path(package, name) as p:
        return p.exists()


def read_text(package: str, resource: str, encoding: str = "utf-8", errors: str = "strict") -> str:
    """Return the decoded string of the resource.

    The decoding-related arguments have the same semantics as those of `pathlib.Path.read_text`.
    """
    with path(package, resource) as p:
        return p.read_text(encoding, errors)



from pyaerocom.aeroval.modelmaps_helpers import _jsdate_list


def test__jsdate_list(data_tm5):
    vals = _jsdate_list(data_tm5)
    assert len(vals) == 12
    assert vals[0] == 1263513600000
    assert vals[-1] == 1292371200000



from copy import deepcopy
from pyaerocom import const

_UNIT_OVERRIDES: None | dict[str, str] = None


def set_unit_overrides(units: dict[str, str]) -> None:
    global _UNIT_OVERRIDES
    units = deepcopy(units)
    if _UNIT_OVERRIDES is None:
        _UNIT_OVERRIDES = units
        return

    if _UNIT_OVERRIDES != units:
        raise ValueError(
            "Units have already been set, and new units dict does not match the old one."
        )

    _UNIT_OVERRIDES = units


def get_standard_unit(var_name: str) -> str:
    """Gets standard unit of AeroCom variable

    Also handles alias names for variables, etc. or strings corresponding to
    older conventions (e.g. names containing 3D).

    Parameters
    ----------
    var_name : str
        AeroCom variable name
    unit_overrides : Optional dict of non-default units, which will be picked first.

    Returns
    -------
    str
        corresponding standard unit

    Units are picked based on the following order (highest priority to lowest):
    1. The units configuration from the user's experiment config.
    2. The default unit configured in variables.ini
    """
    if _UNIT_OVERRIDES is not None:
        if var_name in _UNIT_OVERRIDES:
            return _UNIT_OVERRIDES[var_name]

    return const.VARS[var_name].units



from __future__ import annotations
import datetime
import sys

from pyaerocom.units.molecular_mass import MolecularMass, UnknownSpeciesError, _get_species

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from collections.abc import Iterable


import cf_units
import numpy as np

from .datetime import TsType
from .datetime.time_config import SI_TO_TS_TYPE
from pyaerocom.variable_helpers import get_variable

from typing import TypeVar, overload, NamedTuple
from collections.abc import Callable
from .typing import SupportsMul

__all__ = ["Unit"]

T = TypeVar("T", bound=SupportsMul)


class UnitConversionCallbackInfo(NamedTuple):
    factor: float
    from_aerocom_var: str | None
    from_ts_type: TsType | None
    from_cf_unit: cf_units.Unit
    to_cf_unit: cf_units.Unit


UnitConversionCallbackHandler = Callable[[UnitConversionCallbackInfo], None]


class Unit:
    """Pyaerocom specific encapsulation of cf_units.Unit that extends it
    with additional needed behaviour.

    The first additional behaviour is to handle variables that measure only
    a portion of the real mass. Eg. if concso4 is provided as "ug S/m3", we
    want the mass in terms of SO4, so the values must be scaled up by a
    constant factor MolecularMass("SO4")/MolecularMass("S").

    The second behaviour is adding implicit frequency for rate variables
    and a ts_type. If tstype and aerocom_var are provided in __init__, units
    of the form "mg m-2" will automatically have the temporal frequency
    appended. For instance, assuming tstype='daily', it becomes "mg m-2 d-1"

    Third, cf_units.Unit does not natively support conversion of eg. pd.Series.
    This wrapper allows conversion of any data structure that supports __mul__.
    """

    # If found in the nominator, these are treated as elements for scaling units.
    _TREAT_AS_ELEMENT = ["C", "N", "S"]

    _UALIASES = {
        # mass concentrations
        "ug S m-3": "ug S/m3",
        "ug C m-3": "ug C/m3",
        "ug N m-3": "ug N/m3",
        "ugC/m3": "ug C m-3",
        # deposition rates (implicit)
        ## sulphur species
        "mgS/m2": "mg S m-2",
        "mgSm-2": "mg S m-2",
        ## nitrogen species
        "mgN/m2": "mg N m-2",
        "mgNm-2": "mg N m-2",
        # deposition rates (explicit)
        ## sulphur species
        "mgS/m2/h": "mg S m-2 h-1",
        "mg/m2/h": "mg m-2 h-1",
        "mgS/m**2/h": "mg S m-2 h-1",
        "mgSm-2h-1": "mg S m-2 h-1",
        "mgSm**-2h-1": "mg S m-2 h-1",
        "mgS/m2/d": "mg S m-2 d-1",
        ## nitrogen species
        "mgN/m2/h": "mg N m-2 h-1",
        "mgN/m**2/h": "mg N m-2 h-1",
        "mgNm-2h-1": "mg N m-2 h-1",
        "mgNm**-2h-1": "mg N m-2 h-1",
        "mgN/m2/d": "mg N m-2 d-1",
        ## others
        "MM/H": "mm h-1",
        # others
        "/m": "m-1",
        # attenuated backscatter
        "Mm-1.sr-1": "Mm-1 sr-1",
    }

    def __init__(
        self,
        unit: str,
        calendar: str | None = None,
        *,
        aerocom_var: str | None = None,
        ts_type: str | TsType | None = None,
        **kwargs,
    ) -> None:
        self._origin = str(unit)
        unit = Unit._UALIASES.get(str(unit), str(unit))

        self._species = kwargs.pop("species", None)
        if self._species is None:
            try:
                sp = _get_species(aerocom_var).upper()
            except (UnknownSpeciesError, AttributeError):
                pass
            else:
                self._species = sp

        self._element = None
        if self._species is not None:
            for e in Unit._TREAT_AS_ELEMENT:
                if e in self._origin_nominator:
                    unit = unit.replace(e, "", 1)
                    self._element = e
                    break

        if self._element is not None and self._species is not None:
            try:
                factor = MolecularMass(self._species) / MolecularMass(self._element)
            except ValueError:
                self._species = None
                factor = 1
        else:
            factor = 1

        if factor != 1:
            new_unit = f"{factor} {unit}"
        else:
            new_unit = unit

        if ts_type is not None and aerocom_var is not None and get_variable(aerocom_var).is_rate:
            ends_with_freq = False
            for si_unit in SI_TO_TS_TYPE:
                if unit.endswith(f"/{si_unit}") or unit.endswith(f"{si_unit}-1"):
                    ends_with_freq = True
                    break

            if not ends_with_freq:
                new_unit = f"{new_unit} {TsType(ts_type).to_si()}-1"

        self._aerocom_var = aerocom_var
        self._ts_type = None
        if ts_type is not None:
            self._ts_type = TsType(ts_type)
        self._cfunit = cf_units.Unit(new_unit, calendar=calendar)

    @property
    def origin(self) -> str:
        """
        The original string used to create this Unit.
        """
        return self._origin

    @property
    def _origin_nominator(self) -> str:
        """
        The nominator of the original string used to initialize this Unit instance.
        """
        if "/" in self.origin:
            return self.origin.split("/")[0].strip()

        if "-" in self.origin:
            idx = self.origin.index("-")
            try:
                while self.origin[idx] not in [" ", "."]:
                    idx -= 1
            except IndexError:
                pass
            return self.origin[:idx].strip()

        return self.origin.strip()

    @property
    def _origin_denominator(self) -> str:
        """
        The denominator of the original string used to initialize this Unit instance.
        """
        if "/" in self.origin:
            return self.origin.split("/")[1].strip()

        if "-" in self.origin:
            idx = self.origin.index("-")
            while self.origin[idx] != " ":
                idx -= 1

            return self.origin[idx:].strip()

        return ""

    def is_convertible(self, other: str | Unit) -> bool:
        """
        Return whether this unit is convertible to other. It handles a couple of
        additional cases when checking convertibility, namely:

        - Units that contain elements (eg. kg N m-2) need to have compatible mass ratios:
           - This is assumed to be the case if element and species are the same, the aerocom
           variable is the same (to account for variables that don't have a clear mass ratio —
           eg. wetrdn.)

        :param other: Other Unit.
        """
        if isinstance(other, str):
            other = Unit(other)

        cf_units_only = self._element is None and other._element is None
        compatible_element = (
            (self._element is None)
            or (other._element is None)
            or (self._element == other._element)
        )
        same_species = (self._species is not None and other._species is not None) and (
            self._species == other._species
        )
        same_variable = (self._aerocom_var is not None and other._aerocom_var is not None) and (
            self._aerocom_var == other._aerocom_var
        )

        if cf_units_only:
            return self._cfunit.is_convertible(other._cfunit)

        if not compatible_element:
            return False

        if same_species or same_variable:
            return self._cfunit.is_convertible(other._cfunit)

        return False

    def is_dimensionless(self) -> bool:
        """
        Return whether the unit is dimensionless.
        """
        return self._cfunit.is_dimensionless()

    def is_unknown(self) -> bool:
        """
        Return whether the unit is defined to be an unknown unit.
        """
        return self._cfunit.is_unknown()

    def __str__(self) -> str:
        return self._cfunit.__str__()

    def __repr__(self) -> str:
        result = f"Unit('{self.origin}'"
        if self._aerocom_var is not None:
            result += f", aerocom_var='{self._aerocom_var}'"
        if self._species is not None:
            result += f", species='{self._species}'"
        if self._ts_type is not None:
            result += f", ts_type='{self._ts_type}'"

        result += ")"
        return result

    def __add__(self, other: float) -> Unit:
        return Unit.from_cf_units(self._cfunit.__add__(other))

    def __sub__(self, other: float) -> Unit:
        return Unit.from_cf_units(self._cfunit.__sub__(other))

    def __mul__(self, other: float | str | Unit) -> Unit:
        return Unit.from_cf_units(self._cfunit.__mul__(other))

    def __truediv__(self, other: float | str | Unit) -> Unit:
        return Unit.from_cf_units(self._cfunit.__truediv__(other))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Unit):
            try:
                other = Unit(str(other))
            except ValueError:
                return False

        return self._cfunit.__eq__(other._cfunit)

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    @overload
    def convert(
        self,
        value: int,
        other: str | Self,
        inplace: bool = False,
        *,
        callback: None | UnitConversionCallbackHandler = None,
        **kwargs,
    ) -> float: ...

    @overload
    def convert(
        self,
        value: T,
        other: str | Self,
        inplace: bool = False,
        *,
        callback: None | UnitConversionCallbackHandler = None,
        **kwargs,
    ) -> T: ...

    def convert(
        self,
        value: T,
        other: str | Self,
        inplace: bool = False,
        *,
        callback: None | UnitConversionCallbackHandler = None,
        **kwargs,
    ) -> T:
        """Implements unit conversion to a different unit that should work
        with any data structure that supports __mul__ and / or __imul__.

        :param value: The value to be converted.
        :param other: The unit to which to convert (will be passed to PyaerocomUnit.__init__())
        :param callback: Callback function for eg. logging, defaults to None
            The callback function will receive a NamedTuple with the following keys:
                "factor" - float: The numerical conversion factor used.
                "from_aerocom_var" - str: The aerocom var name.
                "from_ts_type" - str: The ts_type of the from units.
                "from_cf_unit" - str: The base cf_unit converted from.
                "to_cf_unit" - str: The base cf_unit converted to.
        :param kwargs: Will be passed as additional keyword args to PyaerocomUnit.__init__() for 'other'.
        :return: Unit converted data.
        """
        if isinstance(other, str):
            to_unit = Unit(str(other), **kwargs)
        else:
            assert isinstance(other, Unit)
            to_unit = other

        if not self.is_convertible(to_unit):
            raise ValueError(
                f"Unable to convert units. Got incompatible units '{repr(self)}' and '{repr(to_unit)}'."
            )
        to_unit_cf = to_unit._cfunit
        factor = float(self._cfunit.convert(1, to_unit_cf, inplace=False))

        if inplace:
            value *= factor
            result = value
        else:
            result = factor * value

        if isinstance(value, int):
            assert isinstance(result, float)
        else:
            assert type(result) is type(value)

        assert (result is value) == inplace

        if callback is not None:
            info = UnitConversionCallbackInfo(
                factor=factor,
                from_aerocom_var=self._aerocom_var,
                from_ts_type=self._ts_type,
                from_cf_unit=str(self._cfunit),
                to_cf_unit=str(to_unit),
            )
            callback(info)

        return result

    def date2num(
        self, date: datetime.datetime | Iterable[datetime.datetime]
    ) -> float | np.ndarray:
        """Returns the numeric time value calculated from the datetime object using the current calendar and unit time reference.

        :param date: Date to be converted.
        :return:

        See also: https://cf-units.readthedocs.io/en/latest/unit.html#cf_units.Unit.date2num
        """
        return self._cfunit.date2num(date)

    def num2date(
        self,
        time_value: float | np.ndarray,
        only_use_cftime_datetimes: bool = True,
        only_use_python_datetimes: bool = False,
    ) -> datetime.datetime | np.ndarray:
        """
        Returns a datetime-like object calculated from the numeric time value using the current calendar and the unit time reference.

        :param time_value: Time value(s)
        :param only_use_cftime_datetimes:
            If True, will always return cftime datetime objects, regardless of calendar. If False, returns datetime.datetime instances where possible. Defaults to True.
        :param only_use_python_datetimes:
            If True, will always return datetime.datetime instances where possible, and raise an exception if not. Ignored if only_use_cftime_datetimes is True. Defaults to False.
        :return: Datetime or ndarray of datetime.

        See also: https://cf-units.readthedocs.io/en/latest/unit.html#cf_units.Unit.num2date
        """
        return self._cfunit.num2date(
            time_value, only_use_cftime_datetimes, only_use_python_datetimes
        )

    @classmethod
    def from_cf_units(cls, unit: cf_units.Unit) -> Self:
        """
        Initialize from a cf_units.Unit instance.

        :param unit: The input unit.

        :return: The output unit.
        """
        return cls(unit)



import pytest

from pyaerocom.exceptions import VariableDefinitionError
from pyaerocom.varnameinfo import VarNameInfo


@pytest.mark.parametrize(
    "var_name, contains_numbers, contains_wavelength_nm, is_wavelength_dependent",
    [
        ("concpm10", True, True, False),
        ("wetoxn", False, False, False),
        ("od550gt1aer", True, True, True),
    ],
)
def test_varnameinfo(var_name, contains_numbers, contains_wavelength_nm, is_wavelength_dependent):
    var_name_info = VarNameInfo(var_name)
    assert isinstance(var_name_info, VarNameInfo)
    assert var_name_info.contains_numbers == contains_numbers
    assert var_name_info.contains_wavelength_nm == contains_wavelength_nm
    assert var_name_info.is_wavelength_dependent == is_wavelength_dependent
    assert isinstance(str(var_name_info), str)


@pytest.mark.parametrize(
    "var_name, vert_code",
    [("wetoxn", "Surface"), ("od550aer", "Column")],
)
def test_get_default_vert_code(var_name, vert_code):
    var_name_info = VarNameInfo(var_name)
    assert var_name_info.get_default_vert_code() == vert_code


def test_get_default_vert_code_error():
    var_name_info = VarNameInfo("concpm10")
    with pytest.raises(ValueError):
        var_name_info.get_default_vert_code()


@pytest.mark.parametrize(
    "var_name, contains_numbers",
    [
        ("concpm10", True),
        ("od550aer", True),
        ("emitolu", False),
    ],
)
def test_contains_numbers(var_name, contains_numbers):
    var_name_info = VarNameInfo(var_name)
    assert var_name_info.contains_numbers == contains_numbers


@pytest.mark.parametrize(
    "var_name, wavelength",
    [
        ("od550aer", 550),
        ("bsc532aer", 532),
    ],
)
def test_wavelength_nm(var_name, wavelength):
    var_name_info = VarNameInfo(var_name)
    assert var_name_info.wavelength_nm == wavelength


@pytest.mark.parametrize(
    "var_name",
    [
        ("concpm10"),
        ("wetoxs"),
    ],
)
def test_wavelength_nm_errors(var_name):
    var_name_info = VarNameInfo(var_name)
    with pytest.raises(VariableDefinitionError):
        var_name_info.wavelength_nm()


@pytest.mark.parametrize(
    "to_wavelength",
    [
        (350),
        (440),
    ],
)
def test_translate_to_wavelength(to_wavelength):
    od = VarNameInfo("od550aer")
    new_od = od.translate_to_wavelength(to_wavelength)
    assert new_od.wavelength_nm == to_wavelength



from __future__ import annotations

import pytest

from pyaerocom import Colocator, GriddedData, UngriddedData
from pyaerocom.aeroval import EvalSetup
from pyaerocom.aeroval._processing_base import DataImporter, HasColocator, HasConfig
from pyaerocom.aeroval.experiment_output import ExperimentOutput
from pyaerocom.exceptions import EntryNotAvailable


@pytest.fixture(scope="module")
def setup() -> EvalSetup:
    """EvalSetup instance"""
    obs_cfg = dict(
        obs1=dict(obs_id="obs1", obs_vars=["od550aer"], obs_vert_type="Column"),
        obs2=dict(
            obs_id="obs2",
            obs_vars=["od550aer"],
            obs_vert_type="Column",
            diurnal_only=True,
        ),
    )
    return EvalSetup(proj_id="bla", exp_id="blub", obs_cfg=obs_cfg)


@pytest.fixture(scope="module")
def config(setup: EvalSetup) -> HasConfig:
    """HasConfig instance"""
    return HasConfig(setup)


def test_HasConfig_setup(config: HasConfig):
    assert isinstance(config.cfg, EvalSetup)
    assert isinstance(config.exp_output, ExperimentOutput)


def test_HasConfig_raise_exceptions(config: HasConfig):
    assert not config.raise_exceptions


def test_HasConfig_reanalyse_existing(config: HasConfig):
    assert config.reanalyse_existing


@pytest.fixture(scope="module")
def collocator(setup: EvalSetup) -> HasColocator:
    """HasColocator instance"""
    return HasColocator(setup)


@pytest.mark.parametrize("obs_name", [None, "obs1", "obs2"])
def test_HasColocator_get_colocator(collocator: HasColocator, obs_name: str | None):
    col = collocator.get_colocator(obs_name=obs_name)
    assert isinstance(col, Colocator)


def test_HasColocator_get_colocator_error(collocator: HasColocator):
    with pytest.raises(EntryNotAvailable) as e:
        collocator.get_colocator(model_name="mod2")
    assert str(e.value) == "'no such entry mod2'"


@pytest.fixture
def importer(eval_config: dict) -> DataImporter:
    """initialized DataImporter"""
    setup = EvalSetup(**eval_config)
    return DataImporter(setup)


@pytest.mark.parametrize("cfg", ["cfgexp1"])
def test_DataImporter_read_model_data(importer: DataImporter):
    data = importer.read_model_data("TM5-AP3-CTRL", "od550aer")
    assert isinstance(data, GriddedData)


@pytest.mark.parametrize("cfg", ["cfgexp1"])
def test_DataImporter_read_ungridded_obsdata(importer: DataImporter):
    data = importer.read_ungridded_obsdata("AERONET-Sun", "od550aer")
    assert isinstance(data, UngriddedData)



import os
from pyaerocom.variable import Variable
from pyaerocom import const

### Register dummy fraction
variables = {
    "fraction": Variable(var_name="fraction", units="1", description="fraction"),
}

const.register_custom_variables(variables)
OBS_GROUNDBASED = {
    "EBAS-m": dict(
        obs_id="EBASMC",
        web_interface_name="EBAS-m",
        obs_vars=[
            "concwetrdn",
        ],
        obs_vert_type="Surface",
        colocate_time=True,
        ts_type="monthly",
        is_bulk=True,
        bulk_options={
            "concwetrdn": dict(
                vars=["wetrdn", "wetrdnpr"],
                model_exists=False,
                mode="fraction",
                units="mg N m-3",
            )
        },
    ),
    "AERONET-Sun": dict(
        obs_id="AeronetSunV3L2Subset.daily",
        obs_vars=("fraction",),
        obs_vert_type="Column",
        is_bulk=True,
        bulk_options={
            "fraction": dict(
                vars=["od550aer", "od550aer"],
                model_exists=False,
                mode="fraction",
                units="1",
            )
        },
        ts_type="monthly",
    ),
    "AERONET-Sun-exist": dict(
        obs_id="AeronetSunV3L2Subset.daily",
        obs_vars=("abs550aer",),
        obs_vert_type="Column",
        is_bulk=True,
        bulk_options={
            "abs550aer": dict(
                vars=["od550aer", "od550aer"],
                model_exists=True,
                mode="fraction",
                units="1",
            )
        },
        ts_type="monthly",
    ),
}

folder_EMEP = "/lustre/storeB/project/fou/kl/emep/ModelRuns/2022_REPORTING/TRENDS/2013/"


# Setup for models used in analysis
MODELS = {
    "EMEP": dict(
        model_id="EMEP",
        model_data_dir=folder_EMEP,
        gridded_reader_id={"model": "ReadMscwCtm"},
        model_ts_type_read="monthly",
    ),
    "TM5-AP3-CTRL": dict(
        model_id="TM5-met2010_CTRL-TEST",
        model_add_vars=dict(od550csaer=["od550aer"]),
        model_ts_type_read="monthly",
        flex_ts_type=False,
    ),
}

CFG = dict(
    model_cfg=MODELS,
    obs_cfg=OBS_GROUNDBASED,
    json_basedir=os.path.abspath("./data"),
    # coldata_basedir = os.path.abspath('../../coldata'),
    coldata_basedir=os.path.abspath("./coldata"),
    # io_aux_file=os.path.abspath("../../eval_py/gridded_io_aux.py"),
    # if True, existing colocated data files will be deleted
    reanalyse_existing=True,
    only_json=False,
    add_model_maps=False,
    only_model_maps=False,
    clear_existing_json=False,
    # if True, the analysis will stop whenever an error occurs (else, errors that
    # occurred will be written into the logfiles)
    raise_exceptions=True,
    # Regional filter for analysis
    filter_name="ALL-wMOUNTAINS",
    # colocation frequency (no statistics in higher resolution can be computed)
    ts_type="monthly",
    map_zoom="Europe",
    freqs=["daily", "monthly", "yearly"],
    periods=["2010"],
    main_freq="monthly",
    maps_freq="yearly",
    zeros_to_nan=False,
    colocate_time=True,
    resample_how={"vmro3max": {"daily": {"hourly": "max"}}},
    obs_remove_outliers=False,
    model_remove_outliers=False,
    harmonise_units=True,
    regions_how="country",  #'default',#'country',
    annual_stats_constrained=True,
    proj_id="testing_bulk",
    exp_id="bulktest",
    exp_name="Evaluation of EMEP Bulk",
    exp_descr=("Evaluation of EMEP Bulk"),
    exp_pi="Daniel Heinesen",
    public=True,
    # directory where colocated data files are supposed to be stored
    weighted_stats=True,
    var_order_menu=[
        "concwetrdn",
    ],
)



from __future__ import annotations

import logging
import time
from collections.abc import Iterator
from datetime import date, datetime, timedelta
from pathlib import Path

import numpy as np
import pandas as pd

from pyaerocom import const
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.stationdata import StationData
from pyaerocom.ungridded_data_container import UngriddedDataContainer
from pyaerocom.ungriddeddata_structured import UngriddedDataStructured

from .obs import read_csv

AEROCOM_NAMES = dict(
    CO="concco",
    NO2="concno2",
    O3="conco3",
    PM10="concpm10",
    PM25="concpm25",
    SO2="concso2",
)
DATA_FOLDER_PATH = Path("/lustre/storeB/project/fou/kl/CAMS2_83/obs")


logger = logging.getLogger(__name__)


def obs_paths(
    *dates: datetime | date | str,
    root_path: Path | str = DATA_FOLDER_PATH,
    analysis: bool = False,
) -> Iterator[Path]:
    for date in dates:  # noqa: F402
        if isinstance(date, str):
            date = datetime.strptime(date, "%Y%m%d").date()
        if isinstance(date, datetime):
            date = date.date()
        if isinstance(root_path, str):
            root_path = Path(root_path)
        if analysis:
            filename = "%Y%m/obsmacc4verifana_%Y%m%d.csv"
        else:
            filename = "%Y%m/obsmacc4verif_%Y%m%d.csv"
        path = root_path / date.strftime(filename)
        if not path.is_file():
            logger.warning(f"Could not find {path.name}. Skipping {date}")
            continue
        yield path.resolve()


class ReadCAMS2_83(ReadUngriddedBase):
    __version__ = "0.0.0"
    _FILEMASK = ""
    DATA_ID = const.CAMS2_83_NRT_NAME
    DEFAULT_VARS = list(AEROCOM_NAMES.values())
    PROVIDES_VARIABLES = DEFAULT_VARS
    SUPPORTED_DATASETS = [DATA_ID]
    TS_TYPE = "hourly"

    def read(
        self,
        vars_to_retrieve: str | list[str] | None = None,
        files: list[str | Path] | None = None,
        first_file: int | None = None,
        last_file: int | None = None,
    ) -> UngriddedDataContainer:
        """Read observations as ungridded

        :param vars_to_retrieve: pyaerocom-variables to read, defaults to None
            None meaning all default variables
        :param files: files to read, defaults to None
            None meaning all known files in the date-range
        :param first_file: first file to process from from files, defaults to None=0
        :param last_file: last file to process from files, defaults to None=-1
        :raises TypeError: wrong input type
        :return: ungridded data object
        """
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        if isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        if not isinstance(vars_to_retrieve, list):
            raise TypeError(
                f"Unsupported type {type(vars_to_retrieve)}, "
                "vars_to_retrieve supported types are: str | list[str] | None"
            )
        assert all(
            var in self.PROVIDES_VARIABLES for var in vars_to_retrieve
        ), f"this dataset only has {self.PROVIDES_VARIABLES}"

        if files is None:
            files = list(obs_paths(date.today() - timedelta(days=1)))
        if first_file is not None:
            files = files[first_file:]
        if last_file is not None:
            files = files[:last_file]

        start = time.time()
        logger.info("Start read obs")
        # lazy data_iterator returns immediately, unpacked in from_station_data
        data_iterator = self.__reader(vars_to_retrieve, files)
        ungriddeddata = UngriddedDataStructured.from_station_data(
            data_iterator, add_meta_keys=["station_type"]
        )
        logger.info(f"Time needed to convert obs to ungridded: {time.time() - start}s")
        return ungriddeddata

    def read_file(self, filename, vars_to_retrieve=None):
        return self.read(vars_to_retrieve, [filename])

    @classmethod
    def __reader(
        cls, vars_to_retrieve: list[str], files: list[str | Path]
    ) -> Iterator[StationData]:
        logger.info(f"reading {cls.DATA_ID} {vars_to_retrieve=}")
        logger.debug(f"reading from {files=}")
        reverse_aerocom = {v: k for k, v in AEROCOM_NAMES.items()}
        polls = [reverse_aerocom[v] for v in vars_to_retrieve]

        data = pd.concat(read_csv(path, polls=polls) for path in files).drop_duplicates(
            subset=["station", "poll", "time"]
        )
        df: pd.DataFrame
        for station, df in data.groupby("station"):
            logger.info(f"Reading obs for station {station} and variables {vars_to_retrieve}")
            output = dict(
                station_id=station,
                station_name=station,
                station_type=df["station_type"].iloc[0],
                latitude=df["lat"].iloc[0],
                longitude=df["lon"].iloc[0],
                altitude=df["alt"].iloc[0],
                variables=vars_to_retrieve,
                var_info=dict.fromkeys(vars_to_retrieve, dict(units="ug m-3")),
                data_id=cls.DATA_ID,
                ts_type=cls.TS_TYPE,
            )
            df = df.pivot(index="time", columns="poll", values="conc")
            missing = {poll for poll in AEROCOM_NAMES if poll not in df}
            if missing.issubset(vars_to_retrieve):
                logger.debug(f"no relevant data on {station=}, skip")
                continue
            for poll in missing:
                df[poll] = np.nan
            df = df.rename(AEROCOM_NAMES, axis="columns")
            for poll in vars_to_retrieve:
                output[poll] = df[poll]
            yield StationData(**output)



import numpy as np
import pytest

from pyaerocom.config import ALL_REGION_NAME
from pyaerocom.filter import Filter


# TODO: use mark.parametrize for first 2 test functions and call test_Filter
def test_Filter_init():
    assert Filter("EUROPE").name == "EUROPE-wMOUNTAINS"
    assert Filter("noMOUNTAINS-OCN-NAMERICA").name == "NAMERICA-noMOUNTAINS-OCN"


def test_filter_attributes():
    f = Filter(f"{ALL_REGION_NAME}-noMOUNTAINS-LAND")
    assert f.land_ocn == "LAND"
    assert f.region_name == ALL_REGION_NAME
    assert f.name == f"{ALL_REGION_NAME}-noMOUNTAINS-LAND"
    assert not f.region.is_htap()


@pytest.mark.parametrize(
    "filter_name, mean",
    [
        ("EUROPE-noMOUNTAINS-LAND", 0.16616775),
        ("EUROPE-noMOUNTAINS-OCN", 0.1314668),
        ("EUROPE", 0.13605888),
    ],
)
def test_filter_griddeddata(data_tm5, filter_name, mean):
    # use copy so that this fixture can be used elsewhere without being c
    # changed by this method globally
    model = data_tm5.copy()

    f = Filter(filter_name)  # europe only land

    subset = f.apply(model)
    assert np.nanmean(subset.cube.data) == pytest.approx(mean)


@pytest.mark.parametrize(
    "filter_name,num_sites", [(f"{ALL_REGION_NAME}-wMOUNTAINS", 22), ("OCN", 8), ("EUROPE", 7)]
)
def test_filter_ungriddeddata(aeronetsunv3lev2_subset, filter_name, num_sites):
    obs_data = aeronetsunv3lev2_subset

    f = Filter(filter_name)
    num = len(f.apply(obs_data).unique_station_names)
    assert num == num_sites


@pytest.mark.skip(reason=("Need to storee colocateddata object in tmp_dir and provide as fixture"))
def test_filter_colocateddata():
    pass


# =============================================================================
#     data_coloc = pya.colocation.colocate_gridded_gridded(model,
#                                                          sat,
#                                                          ts_type='monthly',
#                                                          filter_name='WORLD-noMOUNTAINS-OCN')
#
#     assert data_coloc.data.sum().values - 111340.31777193633 < 0.001
# =============================================================================



from pyaerocom.aeroval.collections import ObsCollection, ModelCollection
from pyaerocom.aeroval.obsentry import ObsEntry
from pyaerocom.aeroval.modelentry import ModelEntry
import pytest


def test_obscollection_init_and_add_entry():
    oc = ObsCollection()
    oc.add_entry("model1", dict(obs_id="bla", obs_vars="od550aer", obs_vert_type="Column"))
    assert oc

    oc.add_entry(
        "AN-EEA-MP",
        dict(
            is_superobs=True,
            obs_id=("AirNow", "EEA-NRT-rural", "MarcoPolo"),
            obs_vars=["concpm10", "concpm25", "vmro3", "vmrno2"],
            obs_vert_type="Surface",
        ),
    )

    assert "AN-EEA-MP" in oc.keylist()


def test_obscollection_add_and_get_entry():
    collection = ObsCollection()
    entry = ObsEntry(obs_id="obs1", obs_vars=("var1",))
    collection.add_entry("key1", entry)
    retrieved_entry = collection.get_entry("key1")
    assert retrieved_entry == entry


def test_obscollection_add_and_remove_entry():
    collection = ObsCollection()
    entry = ObsEntry(obs_id="obs1", obs_vars=("var1",))
    collection.add_entry("key1", entry)
    collection.remove_entry("key1")
    with pytest.raises(KeyError):
        collection.get_entry("key1")


def test_obscollection_get_web_interface_name():
    collection = ObsCollection()
    entry = ObsEntry(obs_id="obs1", obs_vars=("var1",), web_interface_name="web_name")
    collection.add_entry("key1", entry)
    assert collection.get_web_interface_name("key1") == "web_name"


def test_obscollection_all_vert_types():
    collection = ObsCollection()
    entry1 = ObsEntry(
        obs_id="obs1", obs_vars=("var1",), obs_vert_type="Surface"
    )  # Assuming ObsEntry has an obs_vert_type attribute
    entry2 = ObsEntry(obs_id="obs2", obs_vars=("var2",), obs_vert_type="Profile")
    collection.add_entry("key1", entry1)
    collection.add_entry("key2", entry2)
    assert set(collection.all_vert_types) == {"Surface", "Profile"}


def test_modelcollection_init_and_add_entry():
    mc = ModelCollection()
    mc.add_entry("model1", dict(model_id="bla", obs_vars="od550aer", obs_vert_type="Column"))
    assert mc

    mc.add_entry(
        "ECMWF_OSUITE",
        dict(
            model_id="ECMWF_OSUITE",
            obs_vars=["concpm10"],
            obs_vert_type="Surface",
        ),
    )

    assert "ECMWF_OSUITE" in mc.keylist()


def test_modelcollection_add_and_get_entry():
    collection = ModelCollection()
    entry = ModelEntry(model_id="mod1")
    collection.add_entry("key1", entry)
    retrieved_entry = collection.get_entry("key1")
    assert retrieved_entry == entry


def test_modelcollection_add_and_remove_entry():
    collection = ModelCollection()
    entry = ModelEntry(model_id="obs1")
    collection.add_entry("key1", entry)
    collection.remove_entry("key1")
    with pytest.raises(KeyError):
        collection.get_entry("key1")



from pydantic import ValidationError
import pytest
from pyaerocom.aeroval.glob_defaults import VarWebInfo, VarWebScaleAndColormap


def test_var_web_info():
    var_web_info = VarWebInfo()
    unique_entries = list(set(var_web_info))
    assert len(unique_entries) == len(var_web_info.keys())


def test_var_web_scale_and_colormap():
    vwsc = VarWebScaleAndColormap()
    assert "concso2" in vwsc
    assert "new_var" not in vwsc

    with pytest.raises(FileNotFoundError) as ex:
        VarWebScaleAndColormap("no_such_file")
    assert "no_such_file" in str(ex.value)

    with pytest.raises(ValidationError):
        VarWebScaleAndColormap(**{"bla": ["blub"]})

    vwsc2 = VarWebScaleAndColormap(
        config_file="", **{"new_var": {"scale": [1, 2, 4], "colmap": "my_colormap"}}
    )
    assert "new_var" in vwsc2

    vwsc3 = VarWebScaleAndColormap(new_var2={"scale": [1, 2, 4], "colmap": "my_colormap"})
    assert "new_var2" in vwsc3



from __future__ import annotations

import logging
from importlib import resources
from pathlib import Path
from typing import ClassVar, Any

import yaml
from pydantic import BaseModel, ConfigDict

import pyaerocom as pya

logger = logging.getLogger(__name__)


# TODO Check a validator if extra/kwarg is serializable. Either in json_repr or as a @field_validator on extra

FilterArgs = dict[str, Any]


class PyaroConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    _DEFAULT_CATALOG: ClassVar[Path] = resources.files(pya) / Path(
        "data/pyaro_catalogs/default.yaml"
    )

    ##########################
    #   Model fields
    ##########################

    name: str
    reader_id: str
    filename_or_obj_or_url: Any
    filters: dict[str, FilterArgs]
    name_map: dict[str, str] | None = None  # no Unit conversion option
    post_processing: list[str] | None = None  # List of variables to add through post-processing

    ##########################
    #   Save and load methods
    ##########################

    def json_repr(self):
        return self.model_dump()

    def save(self, path: Path | None = None) -> None:
        name = self.name

        if not path.is_dir():
            raise ValueError(f"{path} must be a directory")
        filename = path / "catalog.yaml" if path is not None else "catalog.yaml"

        if name in self.load_catalog(filepath=filename):
            raise ValueError(f"{name} already exists in catalog {filename}")
        if name in self.load_catalog():
            raise ValueError(f"{name} already exists in default catalog")

        body = {name: self.json_repr()}
        with open(filename, "w") as f:
            yaml.safe_dump(body, f)

    @classmethod
    def from_dict(cls, data: dict):
        return PyaroConfig.model_validate(data)

    @classmethod
    def load(cls, name: str, filepath: Path | None = None):
        if filepath is not None:
            if filepath.is_dir():
                raise ValueError(f"Filepath {filepath} is a directory not a file")
            elif not filepath.exists():
                raise ValueError(f"No catalog with this name exists {filepath}")

        data = cls.load_catalog(filepath)

        if name in data:
            return PyaroConfig.from_dict(data[name])
        else:
            raise ValueError(f"Config {name} was not found in any catalogs")

    @classmethod
    def list_configs(cls, filepath: Path | None = None) -> list[str]:
        data = cls.load_catalog()
        if filepath is not None:
            logger.info(f"Updating with private catalog {filepath}")
            data.update(cls.load_catalog(filepath))

        return list(data.keys())

    @classmethod
    def load_catalog(cls, filepath: Path | None = None) -> dict:
        if filepath is None:
            filepath = cls._DEFAULT_CATALOG

        if not filepath.exists():
            return {}

        with filepath.open() as f:
            data = yaml.safe_load(f)

        return data



from pathlib import Path

import iris
import xarray as xr

import pyaerocom.helpers_landsea_masks as lsm
from pyaerocom import const
from tests.conftest import lustre_avail

TEST_REGIONS = const.HTAP_REGIONS[:2]


def test_available_region_masks():
    assert lsm.available_htap_masks() == [
        "PAN",
        "EAS",
        "NAF",
        "MDE",
        "LAND",
        "SAS",
        "SPO",
        "OCN",
        "SEA",
        "RBU",
        "EEUROPE",
        "NAM",
        "WEUROPE",
        "SAF",
        "USA",
        "SAM",
        "EUR",
        "NPO",
        "MCA",
    ]


@lustre_avail
def test_download_htap_masks():
    paths = [Path(mask) for mask in lsm.download_htap_masks(TEST_REGIONS)]
    assert bool(paths)
    assert all(path.exists() for path in paths)
    assert {path.parts[-3] for path in paths} == {"MyPyaerocom"}
    assert {path.parts[-2] for path in paths} == {"filtermasks"}
    assert [path.name for path in paths] == ["PANhtap.0.1x0.1deg.nc", "EAShtap.0.1x0.1deg.nc"]


def test_get_htap_mask_files():
    paths = [Path(mask) for mask in lsm.get_htap_mask_files(*TEST_REGIONS)]
    assert [path.name for path in paths] == ["PANhtap.0.1x0.1deg.nc", "EAShtap.0.1x0.1deg.nc"]


def test_load_region_mask_xr():
    mask = lsm.load_region_mask_xr(*TEST_REGIONS)
    assert isinstance(mask, xr.DataArray)
    pixels = int(mask.sum())
    assert pixels == 193355


def test_load_region_mask_iris():
    mask = lsm.load_region_mask_iris(*TEST_REGIONS)
    assert isinstance(mask, iris.cube.Cube)
    pixels = int(mask.data.sum())
    assert pixels == 193355


def test_get_mask_value():
    mask = lsm.load_region_mask_xr("WEUROPE")

    assert lsm.get_mask_value(50, 5, mask)
    assert not lsm.get_mask_value(50, 15, mask)


def test_check_all_htap_available():
    should_be = [
        "EAShtap.0.1x0.1deg.nc",
        "EEUROPEhtap.0.1x0.1deg.nc",
        "EURhtap.0.1x0.1deg.nc",
        "LANDhtap.0.1x0.1deg.nc",
        "MCAhtap.0.1x0.1deg.nc",
        "MDEhtap.0.1x0.1deg.nc",
        "NAFhtap.0.1x0.1deg.nc",
        "NAMhtap.0.1x0.1deg.nc",
        "NPOhtap.0.1x0.1deg.nc",
        "OCNhtap.0.1x0.1deg.nc",
        "PANhtap.0.1x0.1deg.nc",
        "RBUhtap.0.1x0.1deg.nc",
        "SAFhtap.0.1x0.1deg.nc",
        "SAMhtap.0.1x0.1deg.nc",
        "SAShtap.0.1x0.1deg.nc",
        "SEAhtap.0.1x0.1deg.nc",
        "SPOhtap.0.1x0.1deg.nc",
        "USAhtap.0.1x0.1deg.nc",
        "WEUROPEhtap.0.1x0.1deg.nc",
    ]

    files = lsm.check_all_htap_available()
    assert sorted(Path(file).name for file in files) == should_be



from __future__ import annotations

import logging
import multiprocessing as mp
from copy import deepcopy
from datetime import date, datetime, timedelta
from pathlib import Path
from typing import Optional

import typer

from pyaerocom import change_verbosity, const
from pyaerocom.io.cams2_83.models import ModelName, RunType
from pyaerocom.io.cams2_83.read_obs import DATA_FOLDER_PATH as DEFAULT_OBS_PATH
from pyaerocom.io.cams2_83.read_obs import obs_paths
from pyaerocom.io.cams2_83.reader import DATA_FOLDER_PATH as DEFAULT_MODEL_PATH
from pyaerocom.scripts.cams2_83.config import CFG
from pyaerocom.scripts.cams2_83.evaluation import EvalType, date_range, runner, runnermedianscores

app = typer.Typer(add_completion=False, no_args_is_help=True)
logger = logging.getLogger(__name__)


def make_model_entry(
    start_date: datetime,
    end_date: datetime,
    leap: int,
    model_path: Path,
    model: ModelName,
    run_type: RunType,
) -> dict:
    return dict(
        model_id=f"CAMS2-83.{model.name}.day{leap}.{run_type.name}",
        model_data_dir=str(model_path.resolve()),
        gridded_reader_id={"model": "ReadCAMS2_83"},
        model_kwargs=dict(
            daterange=[f"{start_date:%F}", f"{end_date:%F}"],
        ),
    )


def make_config(
    start_date: date,
    end_date: date,
    leap: int,
    model_path: Path,
    obs_path: Path,
    data_path: Path,
    coldata_path: Path,
    models: list[ModelName],
    id: str,
    name: str,
    description: str,
    eval_type: EvalType,
    run_type: RunType,
    only_map: bool,
    add_map: bool,
    add_seasons: bool,
    fairmode: bool,
    medianscores: bool,
) -> dict:
    logger.info("Making the configuration")

    if not models:
        models = list(ModelName)

    cfg = deepcopy(CFG)
    cfg.update(
        model_cfg={
            f"{model.name}": make_model_entry(
                start_date,
                end_date,
                leap,
                model_path,
                model,
                run_type=run_type,
            )
            for model in models
        },
        periods=eval_type.periods(start_date, end_date),
        json_basedir=str(data_path),
        coldata_basedir=str(coldata_path),
    )

    if eval_type is not None:
        eval_type.check_dates(start_date, end_date)
        cfg.update(eval_type.freqs_config())

    extra_obs_days = 4 if eval_type in {"season", "long"} else 0
    if run_type != RunType.AN and medianscores:
        obs_dates = date_range(start_date - timedelta(days=1), end_date + timedelta(days=extra_obs_days))
    else:
        obs_dates = date_range(start_date, end_date + timedelta(days=extra_obs_days))
    cfg["obs_cfg"]["EEA"]["read_opts_ungridded"]["files"] = [  # type:ignore[index]
        str(p) for p in obs_paths(*obs_dates, root_path=obs_path, analysis=run_type == RunType.AN)
    ]

    if run_type == RunType.AN:
        cfg.update(forecast_days=1)

    cfg.update(exp_id=id, exp_name=name, exp_descr=description)

    if add_map:
        cfg.update(add_model_maps=True)

    if only_map:
        cfg.update(add_model_maps=True, only_model_maps=True)

    if add_seasons:
        cfg.update(add_seasons=True)    

    if fairmode:
        if eval_type in ["season", "long"]:
            cfg.update(use_cams2_83_fairmode=True)
        else:
            cfg.update(use_fairmode=True)

    return cfg


@app.command()
def main(
    run_type: RunType = typer.Argument(...),
    eval_type: EvalType = typer.Argument(...),
    start_date: datetime = typer.Argument(
        ..., formats=["%Y-%m-%d", "%Y%m%d"], help="evaluation start date"
    ),
    end_date: datetime = typer.Argument(
        ..., formats=["%Y-%m-%d", "%Y%m%d"], help="evaluation end date"
    ),
    leap: int = typer.Argument(0, min=RunType.AN.days, max=RunType.FC.days, help="forecast day"),
    model_path: Path = typer.Option(
        DEFAULT_MODEL_PATH, exists=True, readable=True, help="path to model data"
    ),
    obs_path: Path = typer.Option(
        DEFAULT_OBS_PATH, exists=True, readable=True, help="path to observation data"
    ),
    data_path: Path = typer.Option(
        Path("../../data").resolve(),
        exists=True,
        readable=True,
        writable=True,
        help="where results are stored",
    ),
    coldata_path: Path = typer.Option(
        Path("../../coldata").resolve(),
        exists=True,
        readable=True,
        writable=True,
        help="where collocated data are stored",
    ),
    model: list[ModelName] = typer.Option(
        [],
        "--model",
        "-m",
        case_sensitive=False,
        help="Which model to use. All is used if none is given",
    ),
    id: str = typer.Option(CFG["exp_id"], help="experiment ID"),
    name: str = typer.Option(CFG["exp_name"], help="experiment name"),
    description: str = typer.Option(CFG["exp_descr"], help="experiment description"),
    add_map: bool = typer.Option(False, "--addmap", help="set add_model_maps"),
    only_map: bool = typer.Option(
        False, "--onlymap", help="set add_model_maps and only_model_maps"
    ),
    add_seasons: bool = typer.Option(False, "--addseasons", help="set add_seasons"),
    fairmode: bool = typer.Option(False, "--fairmode", help="set use_fairmode"),
    medianscores: bool = typer.Option(
        False,
        "--medianscores",
        help="If true just the cams2_83-specific statistics are computed, a.k.a. the median scores plots or 'weird' plots, the cache is not cleared and it's assumed that the colocated data is already in place and the regular statistics have already been run",
    ),
    cache: Optional[Path] = typer.Option(
        None,
        help="Optional path to cache. If nothing is given, the default pyaerocom cache is used",
    ),
    dry_run: bool = typer.Option(
        False,
        "--dry-run",
        "-n",
        help="Will only make and print the config without running the evaluation",
    ),
    pool: int = typer.Option(
        1,
        "--pool",
        "-p",
        min=1,
        help="CPUs for reading OBS and running median scores",
    ),
):
    if dry_run:
        change_verbosity(logging.INFO)

    if pool > mp.cpu_count():
        logger.warning(
            f"The given pool {pool} is larger than the maximum CPU count {mp.cpu_count()}."
        )

    cfg = make_config(
        start_date,
        end_date,
        leap,
        model_path,
        obs_path,
        data_path,
        coldata_path,
        model,
        id,
        name,
        description,
        eval_type,
        run_type,
        only_map,
        add_map,
        add_seasons,
        fairmode,
        medianscores
    )
    
    # we do not want the cache produced in previous runs to be silently cleared
    const.RM_CACHE_OUTDATED = False

    analysis = False
    if run_type == RunType.AN:
        analysis = True

    if medianscores:
        if eval_type not in {"season", "long"}:
            logger.error(
                "Median scores calculations are only consistent with a season/long kind of evaluation"
            )
            raise Exception(
                "Median scores calculations are only consistent with a season/long kind of evaluation"
            )
        else:
            logger.info("Special run for median scores only")
            runnermedianscores(cfg, cache, analysis=analysis, dry_run=dry_run, pool=pool)
    else:
        logger.info("Standard run")
        runner(cfg, cache, dry_run=dry_run, pool=pool)


if __name__ == "__main__":
    main()
